;       Last change    August 1993
(in-package color)

; the color specialist implemented here uses Mary Angela Papalaskaris'
; cylindrical color model. There are three axis - hue (goes in a circle
; around the cylinder - 0 to 12), purity (amount of black in color 0-1) and
; dilution (amount of white in color 0-1). Each color has a region defined 
; by a range on each of these axes. Comparisons of colors are done by
; comparing the intervals for each axis.

; Basic colors have the largest ranges, while non-basic colors have smaller
; ranges, usually close to or overlapping a border of the basic colors.
; Hedging a color doubles all its range boundaries  -- actually what it does
; is change the resulting predicate comparison given the color comparisons
; between the two colors - it does not actually change the axis numbers used
; in the comparison. 
; Intensifying operators are now just ignored, but it is possible at some 
; point that they might be changed to narrow the color space of the color
; being intensified.

(tweakable 'epilog::*color-margin* 0.01 
    "Margin of difference allowed for equality" '(float * *))
(tweakable 'epilog::*color-hedged-operators* 
    '(epilog::sort-of epilog::almost)
    "Operators on predicates which can be interpreted as hedged"
    'symbol-list)
(tweakable 'epilog::*color-intensifying-operators* 
    '(epilog::very epilog::extremely epilog::mos)
    "Operators on predicates which can be interpreted as intensifiers"
    'symbol-list)

(defvar *known-colors* nil "List of colors with numbers on them")
(defvar *basic-colors* 
  '(white black blue red orange yellow green purple brown pink grey) 
  "Basic colors - completely partition the color space")

; handle exactly one level
(defun split-ops-and-pred (pred)
  (case (pred-type pred)
	(modified
	 (let ((op (getp pred 'operator))
	       (args (getp pred 'arguments)))
	      (if (or (member op epilog::*color-hedged-operators*)
		      (member op epilog::*color-intensifying-operators*))
		  (if (and (null (cdr args))
			   (equal (getp (car args) 'pred-type) 'constant))
		      (list (car args) op))
		  )))
	(constant (list pred)))
  )

; compare-preds handles not only simple color predicates, but also hedged
; predicates - set up with an operator. So far only sort-of is defined
; to be a hedging operator - all others are considered to leave the 
; predicate in its simple form.

(defun compare-preds (fullpred1 fullpred2 &aux res
				(fpred1 (split-ops-and-pred fullpred1))
				(fpred2 (split-ops-and-pred fullpred2))
			     (top-op1 (cadr fpred1)) 
			     (top-op2 (cadr fpred2))
			     (pred1 (find-symbol (symbol-name (car fpred1)) 'color))
			     (pred2 (find-symbol (symbol-name (car fpred2)) 'color))
			     )
  (when (and pred1 pred2)
  (when (and (or (null top-op1) 
		 (member top-op1 epilog::*color-hedged-operators*)
		 (member top-op1 epilog::*color-intensifying-operators*))
	     (or (null top-op2) 
		 (member top-op2 epilog::*color-hedged-operators*)
		 (member top-op2 epilog::*color-intensifying-operators*)) )
  (setq res
  (if (equal pred1 pred2)
      (if (equal top-op1 top-op2)
          'equivalent
	  (if (member top-op1 epilog::*color-hedged-operators*)
	      'subsumes
	      (if (member top-op2 epilog::*color-hedged-operators*)
		  'subsumed
		  'equivalent)))
      (if (and (member pred1 *known-colors*)
	       (member pred2 *known-colors*))
          (relation pred1 pred2 (member top-op1 epilog::*color-hedged-operators*)
		    (member top-op2 epilog::*color-hedged-operators*))
	  'unknown)))
  (when (and (traced-p 'epilog::color-test) (not (equal res 'unknown)))
	(print-line "    Color specialist: Found relation " res " between " 
		    (print-info fullpred1) " and " (print-info fullpred2)))
  res) ))

; make-color defines a new color to the specialist. It adds the various
; axis ranges, as well as marking the color with the specialist and adding
; the indicator tp.coloring (This isn't really a good way to do things - as
; it is hidden in the code, but it beats putting each color in in three places!)

(defun make-color (c purity-min purity-max dilution-min
			 dilution-max hue-min hue-max &aux
			 color ucolor)
  ; Note - adding colors is NOT retractable - the properties on the colors
  ; themselves could easily be made retractable (use change-property instead
  ; of setp), but *known-colors* could not be reset, which could lead to
  ; problems. Shouldn't be a problem anyway - how often are colors added, never
  ; mind retracted?
  (cond ((equal (symbol-package c) (find-package 'epilog))
	 (setq color (intern (symbol-name c) 'color))
	 (setq ucolor c))
	(t
	 (setq ucolor (intern (symbol-name c) 'epilog))
	 (setq color c))
	)
  (setp color 'PURITY-MIN purity-min)
  (setp color 'PURITY-MAX purity-max)
  (setp color 'DILUTION-MIN dilution-min)
  (setp color 'DILUTION-MAX dilution-max)
  (setp color 'HUE-MIN hue-min)
  (setp color 'HUE-MAX hue-max)
  (create-predicate ucolor 'epilog::color-specialist)
  (add-indicate ucolor 'epilog::tp.coloring)
  (pushnew color *known-colors*)
  )

; rmod is a mod function for real numbers
(defun rmod (r i &aux
	       (ri (truncate r))
	       (rr (- r ri)))
  (+ (mod ri i) rr))

; lshift handles boundary conditions (0 and 12)
(defun lshift (x amount)
  (rmod (- (+ x 12) amount) 12)
  )

; eqtest tests to see if boundaries are equal
(defun eqtest (a b)
  (or (< (abs (- a b)) epilog::*color-margin*)
      (< (- 12 (abs (- a b))) epilog::*color-margin*))
  )

; handles the interval comparisons when none of the
; intervals crosses the origin
(defun rel-1 (xmin xmax ymin ymax)
  (cond ((or (eqtest xmin ymax) (eqtest ymin xmax))
	 'next)
	((or (> xmin ymax) (> ymin xmax))
	 'sep)
	((and (<= xmin ymin) (<= ymax xmax))
	 'left)
	((and (<= ymin xmin) (<= xmax ymax))
	 'right)
	(t 'partial)
	)
  )

; rel-interval computes relation between two intervals (purity, dilution or hue)
; resulting relation one of: equal, left, right, sep, partial

(defun rel-interval (xmin xmax ymin ymax) 
  (cond ((and (eqtest xmin ymin) (eqtest xmax ymax))
	 'equal)
	((eqtest xmin xmax)
	 'left)
	((eqtest ymin ymax)
	 'right)
	((< xmin xmax)
	 (if (< ymin ymax)
	     (rel-1 xmin xmax ymin ymax)
	     (let ((samount ymin))
		  (setq xmin (lshift xmin samount))
		  (setq xmax (lshift xmax samount))
		  (setq ymin (lshift ymin samount))
		  (setq ymax (lshift ymax samount))
		  (if (eqtest xmin xmax)
		      'left
		      (rel-1 xmin xmax ymin ymax))
		  )))
	(t ; shifting of x
	   (let ((samount xmin))
		(setq xmin (lshift xmin samount))
		(setq xmax (lshift xmax samount))
		(setq ymin (lshift ymin samount))
		(setq ymax (lshift ymax samount))
		(if (> ymin ymax)
		    (if (> ymax xmax)
			'right
			(if (and (eqtest ymax xmin) (> ymin xmax))
			    'next
			    'partial))
		    (if (eqtest ymin ymax)
			'right
			(rel-1 xmin xmax ymin ymax))
		    )
		)
	   ))
  )

; centre tests two intervals for common endpoints - returns t if neither is
; common
(defun centre (xmin xmax ymin ymax)
  ; tests for common endpoints
  ; returns true when neither is common
  (not (or (<= (abs (- xmin ymin)) epilog::*color-margin*)
	   (<= (abs (- xmax ymax)) epilog::*color-margin*)))
  )

; color-relation computes the relation between two colors by considering
; the relations between their ranges on each of the purity, dilution and
; hue axes.

(defun color-relation (color1 color2 &aux
			      (pmin1 (getp color1 'PURITY-MIN))
			      (pmax1 (getp color1 'PURITY-MAX))
			      (pmin2 (getp color2 'PURITY-MIN))
			      (pmax2 (getp color2 'PURITY-MAX))
			      (rel-p (rel-interval pmin1 pmax1 pmin2 pmax2))
			      (dmin1 (getp color1 'DILUTION-MIN))
			      (dmax1 (getp color1 'DILUTION-MAX))
			      (dmin2 (getp color2 'DILUTION-MIN))
			      (dmax2 (getp color2 'DILUTION-MAX))
			      (rel-d (rel-interval dmin1 dmax1 dmin2 dmax2))
			      (hmin1 (getp color1 'HUE-MIN))
			      (hmax1 (getp color1 'HUE-MAX))
			      (hmin2 (getp color2 'HUE-MIN))
			      (hmax2 (getp color2 'HUE-MAX))
			      (rel-h (rel-interval hmin1 hmax1 hmin2 hmax2))
			      (relset (list rel-p rel-d rel-h))
			      )
  (when (traced-p 'epilog::color-details)
	(print-line "    Color Specialist: Comparing " color1 " with " color2 )
	(print-line "       purity relation " rel-p)
	(print-line "       dilution relation " rel-d)
	(print-line "       hue relation " rel-h))

  (cond ((or (member 'sep relset) 
	     (every #'(lambda (a) (equal a 'next)) relset))
	 'apart)
	((member 'next relset)
	 'adjacent)
	((member 'partial relset)
	 'overlap)
	(t (let ((c (and (centre pmin1 pmax1 pmin2 pmax2)
			 (centre dmin1 dmax1 dmin2 dmax2)
			 (centre hmin1 hmax1 hmin2 hmax2))))
		(cond ((every #'(lambda (a) (member a '(left equal))) relset)
                       (if c 'leftcentre 'leftborder))
		      ((every #'(lambda (a) (member a '(right equal))) relset)
                       (if c 'rightcentre 'rightborder))
		      (t ; error!
			 'unknown)
		      )))
	)
  )

; reverse-reln reverses the relationship in the case where rightcenter or rightborder
; is the relation, or the relation was overlaps or adjacent with the only
; basic color term as the second one
(defun reverse-reln (a b r &aux 
		  (newrel (final-relation a b r)))
  (if (equal newrel 'subsumes)
      'subsumed
      newrel)
  )

; relation converts the relation between the two basic colors into a
; relationship usable by the main system, and conisders whether or not
; the predicates were hedged as well.

(defun relation (color1 color2 &optional (hedged1 nil) (hedged2 nil) &aux
			(predrel (color-relation color1 color2))
			(b1 (member color1 *basic-colors*))
			(b2 (member color2 *basic-colors*)))
  (when (traced-p 'epilog::color-details)
	(print-line "    Color specialist: Found color relation " predrel
		       " between " color1 (if hedged1 " (hedged) " " ") 
		       "and " color2 (if hedged2 " (hedged)" " "))
	)
  (cond ((and (member predrel '(adjacent overlap))
	      (or (and b2 (not b1))
		  (and (equal b1 b2) (not hedged2) hedged1)))
	 (reverse-reln hedged2 hedged1 predrel))
	(t (final-relation hedged1 hedged2 predrel))
	))

; final-relation determines the relation to return to the specialist interface
(defun final-relation (hedged1 hedged2 predrel)
  (case predrel
	(overlap 
	   (if (not hedged1)
	       'unknown
	       'subsumes))
	(adjacent 
           (if (and hedged1 hedged2)
	       'unknown
               'disjoint))
	(apart 'disjoint)
	(leftborder 
           (if (and (not hedged1) hedged2)
	       'unknown
	       'subsumes))
	(rightborder 
           (if (and hedged1 (not hedged2))
	       'unknown
	       'subsumed))
	(leftcentre 'subsumes)
	(rightcentre 'subsumed)
	(otherwise 'disjoint)
	)
  )
