;       Last change    August 1993
(in-package epilib)
;   Hierarchy routines

; There are two kinds of hierarchies - overlap and exclusion;
; All hierarchies are considered exclusion unless set-hier-type is used to
; indicate otherwise.
; Sibling nodes in an exclusion hierarchy are considered incompatible; in
; an overlap hierarchy they are not - inference must be used to determine
; incompatibility.
; A preorder numbering scheme is used to number the nodes on the hierarchy so 
; that relationships between nodes in the same hierarchy can be determined
; in constant time. Numbering is done is a depth first manner, with each
; node getting a number iof its own, and the maximum number of any subnode
; beneath it.

; Any number of type hierarchies may exist at one time. They may overlap
; with specific type predicates. A list of these connections is kept for each
; hierarchy and used to determine the relationships between predicates on
; different hierarchies. The same predicate may only appear once in each
; hierarchy however.

; In addition to the type hierarchies, there must also be a topic hierarchy
; (and only one).
; The topic hierarchy must be called tp.topics
;   special entries of topic hierarchy:
;         tp.type - concept types classified here
;         tp.spec - instances of types classified here
;         tp.content - _ propositions classified here
;         tp.description - **, * and @ episodes classified here
;         tp.mental-attitude - modal subnets classified here
;         tp.role - part and role subnets classified here
;         tp.mental-attitude-object - modal propositions about facts

; Indicator links from the predicates indicate the topics on the topic hierarchy
; These are stored in a list associated with each predicate
; The list consists of pairs of topic and argument number. When a proposition
; containing that predicate is classified, it will be done so only for the 
; argument listed with the topic indicator. If no argument number is given, 
; it will be classified under all arguments.

; The following routines are available here:
;   For the hierarchies themselves:
;     add-hier hier-name main-node sub-node1 sub-node2 ...
;        adds a new hierarchy for main-node if it doesn't already belong
;        to one, and adds sub-node1 ... under it regardless
;     set-hier-type hier-name type
;        where type is one of exclusion or overlap
;        sets the hierarchy's type
;     print-hier brief hier-name 
;        prints the hierarchy, indenting to show the structure
;        the preorder numbering is also printed
;     number-hier 
;        puts the preorder numbering on all of the hierarchies
;     hier-rel node1 node2
;        retuirns the relationship between the two nodes
;        (disjoint, subsumes, subsumed, equivalent, unknown)

;   For topic indicators
;     add-indicate predicate (indicator1 #) indicator2  ...
;        adds indicator links to a predicate
;     add-topic topic predicate1 (optional #) predicate2 (optional #) ...
;        adds indicator links to topic for each predicate

;   General:
;     print-pred pred
;        prints information about a predicate, including which hierarchy
;        it belongs to (for type), and what topics it indicates

; Global variables and constants:
;     *hierarchies* 
;          contains the root name of each distinct type hierarchy
;     *all-hierarchies* 
;          contains the root name of each distinct hierarchy

(traceable 'hier "Shows connection of hierarchies to ENTITY")

(tweakable '*remainder-indicators* '(remainder other rest)
  "List of allowed endings for remainder indicators in exhaustive hierarchies"
  'symbol-list)
(tweakable '*familiar-parts*
  '(head face ear eye nose mouth tongue tooth neck torso leg arm wing tail 
    finger toe knee elbow flipper fin body wrist ankle abdomen chest back trunk
    brain skull)
  "List of familiar parts that should be known about"
  'symbol-list)
(tweakable '*part-hierarchy-types* '(form function state)
  "Part hierarchy types" 'symbol-list)

(defvar *hierarchies* nil "Keeps track of roots of all type hierarchies")
(defvar *all-hierarchies* nil "Keeps track of roots of all hierarchies")
(defvar *hier-change* nil "Notes whether change made to any hierarchy")
(defvar *allowed-part-hiers* nil)

(defun add-super (node hier super &aux
		       (oldsup (getp node 'SUPER))
		       oldp)
  (if (setq oldp (assoc hier oldsup))
      (let ()
	   (unless (equal (cdr oldp) super)
		 (print-error node " already exists on hierarchy " 
				hier " - not added")
		 )
	   oldsup)
      (acons hier super oldsup)
      )
  )

(defun add-sub (node hier sub &aux
		       (oldsub (getp node 'SUB))
		       oldpiece
		       )
  (if (setq oldpiece (assoc hier oldsub))
      (nondest-rplacd oldsub hier (append (cdr oldpiece) sub))
      (acons hier sub oldsub)
      )
  )

(defun add-assoc-prop (node hier piece prop &aux
		       (sub (list piece))
		       (oldsub (getp node prop))
		       oldpiece
		       )
  (if (setq oldpiece (assoc hier oldsub))
      (nondest-rplacd oldsub hier (append (cdr oldpiece) sub))
      (acons hier sub oldsub)
      )
  )

(defun common-hier (nlist &aux int)
  (dolist (n nlist)
	  (if (equal n (car nlist))
	      (setq int (getp n 'HIER))
	      (setq int (intersection int (getp n 'HIER)))
	      )
	  )
  int)

(defun get-subs-all (node &aux
			   (subs (get-all-sub node)))
  (remove-duplicates (apply 'append subs (mapcar 'get-subs-all subs)))
  )

(defun get-sup-all (node &aux
			   (sup (get-all-super node)))
  (remove-duplicates (apply 'append sup (mapcar 'get-sup-all sup)))
  )

(defun common-subtype (nodea nodeb &aux
			     int subsa subsb)
  (setq subsa (get-subs-all nodea))
  (setq subsb (get-subs-all nodeb))
  (when (setq int (intersection subsa subsb))
	(print-error nodea " and " nodeb " have common subtype(s) " int )
	(return-from common-subtype t))
  nil)

(defun subtype (nodea nodeb &optional not-error &aux 
		      (subsb (get-all-sub nodeb)))
  (cond ((null subsb) nil)
	((member nodea subsb :test 'equal)
	 (unless not-error (print-error nodea " is a subtype of " nodeb ))
	 t)
	(t (dolist (s subsb)
		   (if (subtype nodea s not-error)
		       (return-from subtype 
				    (let ()
					 (unless not-error
					 (print-line "    so " nodea 
						     " is a subtype of " nodeb))
					 t)))))
	)
  )

(defun connection-node (hier nodea &aux)
  (dolist (c (getp hier 'CONNECTIONS))
	  (if (member nodea (cdr c) :test 'equal)
	      (return-from connection-node t)))
  )

(defun get-conn-parents (node &aux par
			      (sup (get-all-super node)))
  (setq par
  (remove nil (mapcar #'(lambda (a) 
				(if (> (list-length (getp a 'HIER)) 1)
				    a
				    nil)) sup)))
  (union par (apply 'append (mapcar 'get-conn-parents sup)))
  )

(defun disjoint-hier-nodes (nodea nodeb &aux
				 conn1 conn2 lcp sup1 sup2)
  ; if they have a common parent anywhere in an exlcusion hierarchy - disjoint
  ; get lowest common parent - if in same hierarchy as they are, or not a
  ; connection node, disjoint
  (setq sup1 (cons nodea (get-sup-all nodea)))
  (setq sup2 (cons nodeb (get-sup-all nodeb)))
  (setq conn1 (remove nil (mapcar #'(lambda (a) 
				   (if (> (list-length (getp a 'HIER)) 1) 
				       a 
				       nil)) sup1)))
  (setq conn2 (remove nil (mapcar #'(lambda (a) 
				   (if (> (list-length (getp a 'HIER)) 1) 
				       a 
				       nil)) sup2)))
  (setq lcp (car (intersection sup1 sup2)))

  (when (and lcp (not (equal lcp nodea)) (not (equal lcp nodeb)))
	(dolist (h (getp lcp 'HIER))
		(when (and (equal (getp h 'HIER-TYPE)  'EXCLUSION)
			   (or (and (member h (getp nodea 'HIER))
				    (member h (getp nodeb 'HIER)))
			       ))
		      (print-error nodea " disjoint from " nodeb)
		      (return-from disjoint-hier-nodes t))
		)
	; try for connection parents
	(dolist (c1 (remove nodea (remove-duplicates conn1)))
		(when (disjoint-hier-nodes c1 nodeb)
		      (print-line "   so " nodea " disjoint from " nodeb)
		      (return-from disjoint-hier-nodes t)))
	(dolist (c2 (remove nodeb (remove-duplicates conn2)))
		(when (disjoint-hier-nodes nodea c2)
		      (print-line "   so " nodea " disjoint from " nodeb)
		      (return-from disjoint-hier-nodes t)))
	)
  nil)

(defun incompatible-subtype (nodea nodeb &aux
				   (subsb (get-all-sub nodeb)))
  (dolist (s subsb)
	  (when (disjoint-hier-nodes nodea s)
		 (print-error nodea " incompatible with subtype "
			       s " of " nodeb)
		 (return-from incompatible-subtype t))
	  )
  (dolist (s subsb)
	  (if (incompatible-subtype nodea s)
	      (return-from incompatible-subtype t)))
  nil)

(defun hier-consistency (hier node subs &aux checksubs)
  ; numbering has probably not been done, so check w/o numbers.
  ; also, checking is slightly different - unknown isn't enough
  ; to determine whether it is ok to add

  ;need to check:
  ;  are node and all subs already part of the same hier? If so dont add again.
  (unless (getp hier 'part-hier)
  (when (common-hier (cons node subs))
	(warn-error "   " node " and " subs 
			" are already on a common hierarchy ")
	(return-from hier-consistency nil))
	  )

  ;  for each sub
  ;    check to see if sub is disjoint from node
  ;    check if sub has subtype incompatible with node
  (dolist (i subs)
	  (if (or (subtype node i)
		  (if (subtype i node t)
		      nil
	              (disjoint-hier-nodes node i) )
		  (incompatible-subtype node i))
	      (return-from hier-consistency nil))
	  )

  ;  if exclusion hier 
  ;    make sure no two subs have common subconcept
  ;    make sure no sub is a subconcept of another sub (also for overlap?)
  ; note - include subs already stored in these tests too - may be
  ; incrementally added

  (when (equal (getp hier 'HIER-TYPE) 'EXCLUSION)
	(setq checksubs (get-sub hier node))
	(dolist (i subs)
		(do ((j (car checksubs) (car checksubs)))
		    ((null j))
		    (setq checksubs (cdr checksubs))
		    (unless (equal i j)
			    (if (or (common-subtype i j)
				    (subtype i j)
				    (subtype j i))
		 	     (return-from hier-consistency 
			      (print-line "    so " i 
					  " is already a subtype of " node)))
			    ))
		(setq checksubs (append checksubs (list i)))
		)
	)

  t)

(defun part-type (hier) (or (getp hier 'part-hier) (getp hier 'part-type-hier)))

;; add-hier adds a new hierarchy node, or adds to an existing hierarchy
;; arguments are of the form 
;;       hier-name node subnode1 subnode2 ...
;; if node does not already belong to a hierarchy, it is made the root of
;; a new hierarchy, and added to *hierarchies*
(defun add-hier (hier &rest gargs &aux
				  (noclean (member '-noclean gargs))
				  (args (if noclean (remove '-noclean gargs) gargs))
				  (root (car args))
				  (subs (cdr args))
				  usesubs)

  ;; if in gargs there is the -noclean symbol then it calls
  ;; rewind-question before adding the new hierarchy. FM
  (unless noclean
;     (print-warn "call to add-hier with -noclean flag. Is what you want?")
;     (break)
    (rewind-question)
    )
  
  (when (and subs (> *specialist-entry-effort* 0) (not (part-type hier)))
    (unless (hier-consistency hier root subs)
      (print-line "   " subs " not added to " root)
      (return-from add-hier nil)
      )
    )

  (change-property hier 'hier-change t)

  (unless (getp hier 'ROOT)
    ;; an hierarchy has only one root
    (change-property hier 'ROOT root)
    ;; set up new hierarchy
    (when (traced-p 'hier)
      (print-line "Starting new hierarchy " hier " with root " root))

    ;; the node root can be use in multiple hierarchies
    (change-property root 'HIER (union (getp root 'HIER) (list hier)))
    (push hier *all-hierarchies*)
    
    ;; anything on a hierarchy is a type predicate - make them all
    ;; indicate tp.type  (except topics)
    ;; hierarchies are by default type hierarchies - others are
    ;; noted by PRED-HIER on the root
    (unless (equal hier 'TP.TOPICS)
      (unless (getp hier 'PRED-HIER)
	(push hier *hierarchies*))
      (if (getp hier 'PRED-HIER)
	  ;; changed from hier to root. FM
	  (safe-create-predicate root 'hier-specialist noclean)
	;; changed from hier to root. FM
	(safe-create-predicate root 'type-specialist noclean))
      (if (part-type hier)
	  ;; changed from hier to root. FM
	  (safe-create-predicate root 'part-specialist noclean))
      (let ((oldi (getp hier 'INDICATORS)))
	(unless (or (getp hier 'PRED-HIER)(member 'TP.TYPE oldi))
	  (change-property hier 'INDICATORS (cons 'TP.TYPE oldi))))
      )
    
    ;; hierarchy type is by default exclusion - may be set to overlaps 
    ;; before or after asserting the rest of the hierarchy information
    (unless (getp hier 'HIER-TYPE)
      (change-property hier 'HIER-TYPE 'EXCLUSION))
  
    (dolist (h (remove hier (getp root 'HIER)))
      ;; add to connection list
      (unless (and (getp hier 'part-hier) (getp h 'part-hier)
		   (not (equal (getp hier 'root) (getp h 'root))))
	(change-property h 'CONNECTIONS 
			 (add-assoc-prop h hier root 'CONNECTIONS))
	(change-property hier 'CONNECTIONS 
			 (add-assoc-prop hier h root 'CONNECTIONS)) )
      )
    )

  (when (not (member hier (getp root 'HIER)))
    (print-error root " is not on hierarchy " hier 
		 " - incorrect arguments to add-hier")
    (return-from add-hier nil)
    )
  
  (when (and (not (getp hier 'part-hier)) (not (getp hier 'part-type-hier))
	     (or (member-if #'(lambda (a) (part-type a))
			    (getp root 'hier))
		 (member-if #'(lambda (a) (member-if #'(lambda (b) 
							 (and (part-type b) (not (equal a (getp b 'root)))))
						     (getp a 'hier)))
			    subs) ))
    (set-hier-type hier 'part-type-hier t))

  ;; now add subs to root, and make sure all know which hierarchy they
  ;; belong to
  (dolist (s subs)
    (if (member hier (getp s 'HIER))
	(print-error s " already exists on hierarchy " hier
		     " - not added below " root)
      (let ()
	(setq usesubs (append usesubs (list s)))
	(change-property s 'SUPER (add-super s hier root))
	(change-property s 'HIER (union (list hier) (getp s 'HIER)))
	
	(unless (getp s 'pred-interest)
	  (change-property s 'pred-interest (getp root 'pred-interest)))
	(unless (getp s 'sort)
	  (change-property s 'sort (getp root 'sort)))
	(if (and (getp hier 'pred-hier) (getp root 'indicators))
	    (change-property s 'indicators
			     (union (getp s 'indicators) 
				    (getp root 'indicators) :test 'equal)))
	
	;; anything on a hierarchy is a type predicate - make them all
	;; indicate tp.type  (except topics)
	(unless (equal hier 'TP.TOPICS)
	  (if (getp hier 'PRED-HIER)
	      (safe-create-predicate s 'hier-specialist noclean)
	    (safe-create-predicate s 'type-specialist noclean)
	    )
	  (if (part-type hier)
	      (safe-create-predicate s 'part-specialist noclean))
	  (let ((oldi (getp s 'INDICATORS)))
	    (unless (or (getp hier 'PRED-HIER) (member 'TP.TYPE oldi))
	      (if (getp hier 'part-hier)
		  (change-property s 'INDICATORS (cons 'TP.TYPE oldi))
		(change-property s 'INDICATORS (cons 'TP.TYPE oldi))))) )
	(dolist (h (remove hier (getp s 'HIER)))
	  ;; add to connection list
	  (unless (and (getp hier 'part-hier) (getp h 'part-hier)
		       (not (equal (getp hier 'root) (getp h 'root))))
	    (change-property h 'CONNECTIONS 
			     (add-assoc-prop h hier s 'CONNECTIONS))
	    (change-property hier 'CONNECTIONS 
			     (add-assoc-prop hier h s 'CONNECTIONS)) )
	  ) 
	)))
  (when (and usesubs (or (getp hier 'part-hier) (getp hier 'part-type-hier)))
    (unless (getp root 'pred-parts)
      (pred-info root t))
    (dolist (s usesubs)
      (unless (getp s 'pred-parts)
	(pred-info s t)))
    )
  (when usesubs 
    (if (getp hier 'part-hier) 
	(apply 'add-hier (append (list 'part 'part) 
				 (set-difference usesubs (get-sub 'part 'part)))))
    (change-property root 'SUB (add-sub root hier usesubs)))
  )

;; add-hier used to call create-predicate directly, but that was bad:
;; add-hier can be called in the middle of answering a question (with
;; the -noclean flag), and create-predicate calls rewind-question,
;; which causes the question answering to fail.

;; I think in the cases when the -noclean flag is passed,
;; create-predicate doesn't really need to be called.  If I'm wrong,
;; this function will signal an error. -ANK
(defun safe-create-predicate (p spec noclean)
  (if noclean
      (assert (and (getp p 'predicate)
		   (member spec (getp p 'specialists))))
    (create-predicate p spec)))

; set-hier-type allows the type of a hierarchy to be changed - type must be
; exclusion or overlap
; when initially set up, a hierarchy has exclusion as its type.
; set-hier-type may be used before other hierarchy information is asserted, or
; after.

(defun set-hier-type (hier type &optional noclean)
  (unless noclean (rewind-question))

  ; type must be either exclusion or overlap
  (cond ((member type '(EXCLUSION OVERLAP))
         (change-property hier 'HIER-TYPE type))
	((equal type 'TYPE) 
	 (change-property hier 'PRED-HIER nil)
	 (unless (hier-property hier 'exhaustive)
		 (add-hier-property hier 'non-exhaustive t))
	 )
	((equal type 'PRED-HIER) 
	 (change-property hier 'PRED-HIER t)
	 (unless (hier-property hier 'exhaustive)
		 (add-hier-property hier 'non-exhaustive t))
	 )
	((equal type 'PART-TYPE-HIER) 
	 (change-property hier 'PART-TYPE-HIER t)
	 (unless (hier-property hier 'exhaustive)
		 (add-hier-property hier 'non-exhaustive t))
	 )
	((equal type 'PART-HIER) 
	 (change-property hier 'PART-HIER t)
	 (unless (hier-property hier 'non-exhaustive)
		 (add-hier-property hier 'exhaustive t))
	 )
	((member type *part-hierarchy-types*)
	 (change-property hier 'part-type type)
	 )
	((equal type 'EXHAUSTIVE)
	     (change-property hier 'PROPERTIES 
			    (union (list type) 
			    (remove 'non-exhaustive (getp hier 'properties)))) )
	((equal type 'NON-EXHAUSTIVE)
	     (change-property hier 'PROPERTIES 
			    (union (list type) 
			    (remove 'exhaustive (getp hier 'properties)))) )
	(t (print-error "Hierarchy type must be either EXCLUSION or OVERLAP"
			  *nl* "    or TYPE or PRED-HIER"))
      ))

(defun hier-property (hier property 
			   &aux (props (getp hier 'properties)))
  (cond ((equal property 'exhaustive)
	 (if (member 'non-exhaustive props)
	     nil
	     (member 'exhaustive props)))
	((equal property 'non-exhaustive)
	 (if (member 'exhaustive props)
	     nil
	     (member 'non-exhaustive props)))
	)
  )

(defun add-hier-property (hier property &optional noclean)
  (unless noclean (rewind-question))

  (cond ((equal property 'EXHAUSTIVE)
	     (change-property hier 'PROPERTIES 
			    (union (list property) 
			    (remove 'non-exhaustive (getp hier 'properties)))) )
	((equal property 'NON-EXHAUSTIVE)
	     (change-property hier 'PROPERTIES 
			    (union (list property) 
			    (remove 'exhaustive (getp hier 'properties)))) )
	((member property 
		 '(exclusion overlap type pred-hier part-type-hier part-hier))
	 (set-hier-type hier property t))
	((member property *part-hierarchy-types*)
	 (set-hier-type hier property t))
	(t (print-error "Property " property " is not recognized")))
  )

(defun add-num (hier node num prop &aux
		     (oldmax (getp node prop)))
  (if (assoc hier oldmax)
      (nondest-rplacd oldmax hier num)
      (acons hier num oldmax))
  )

; number all the hierachies 
(defun number-hier (&optional hier &aux
			      root)
  (cond (hier
	 (setq root (getp hier 'ROOT))
	 (change-property root 'MAX  
	       (add-num hier root (number-hier-node hier root 1) 'MAX))
	 (change-property hier 'hier-change nil)
	 (if (getp hier 'part-hier) 
	     (calc-part-totals hier)
	     (check-root-connect hier))
	 )
	(t (dolist (h *all-hierarchies*)
	     (number-hier h))
	   (when (getp 'entity 'hier-change) (number-hier 'entity))
	   ))
  )

(defun check-root-connect (hier &aux r
				     (root (getp hier 'root)))
  (unless (or (equal hier 'entity)
	      (equal hier 'tp.topics)
	      (getp hier 'pred-hier) 
	      (getp hier 'part-hier)
	      (getp hier 'part-type-hier)
	      (and (setq r (hier-rel root 'entity))
		   (not (equal r 'unknown))))
    (when (traced-p 'hier)
      (print-line "Adding " root " to ENTITY hierarchy"))
    (add-hier 'entity 'entity root '-noclean)
    )
  )

; number a particular node and all below it - returns the max
(defun number-hier-node (hier node id &optional (level 1) &aux subs old ptype)
  (change-property node 'ID (add-num hier node id 'ID))
  (setq subs (cdr (assoc hier (getp node 'sub))))
  (setq ptype (cond ((getp hier 'part-type)) (t 'form)))
  (mapc #'(lambda (h) 
		  (if (and (member h *familiar-parts*)
			   (or (null (setq old 
				   (cdr (assoc ptype (getp h 'normal-level)))))
			       (> level old)))
		      (change-property h 'normal-level 
				 (nondest-rplacd (getp h 'normal-level) ptype level)))
		  (setq id (number-hier-node hier h (+ 1 id) (+ level 1))) 
		  (change-property h 'MAX (add-num hier h id 'MAX))
		  )
	subs)
  id)

(set-display-function 'hier 'hier-print "Prints type hierarchies")

(defun hier-print (brief &rest nodes)
  (unless nodes
	  (setq nodes (remove 'tp.topics *all-hierarchies* :test 'equal)))
  (dolist (n nodes)
	  (print-hier brief n))
  )

; print-hier prints out a hierarchy, indenting to show the different levels
(defun print-hier (brief hier &aux  hiers
				    (root (getp hier 'ROOT)))
  (unless root 
    (print-line hier "   Hierarchies: " (setq hiers (getp hier 'hier)))
    (dolist (h hiers)
      (if (getp h 'hier-change) (number-hier h))
      (print-hier-node h hier)
      (print-line))
    (return-from print-hier))
  (if (getp hier 'hier-change)
      (number-hier hier))
  (print-line)
  (print-it hier "   Root: " (getp hier 'ROOT) 
	    "   Hierarchy type: " (getp hier 'HIER-TYPE))
  (when (getp hier 'part-hier) (print-it "    PART HIERARCHY"))
  (when (getp hier 'part-type-hier) (print-it "    PART TYPE HIERARCHY"))
  (when (getp hier 'properties) (print-line "        PROPERTIES: " 
					    (getp hier 'properties)))
  (when (getp hier 'part-type) 
    (print-it "    " (getp hier 'part-type) " part hierarchy"))
  (when (getp hier 'CONNECTIONS)
    (print-line "  Connections to other hierarchies:")
    (dolist (c (getp hier 'CONNECTIONS))
      (print-line "      Hier: " (car c) 
		  "     Predicates: " (cdr c))
      ))
  (terpri)
  (unless brief
    (if (getp hier 'part-hier)
	(print-part-hier-node hier root '(1 1))
      (print-hier-node hier root)) )
  (print-line)
  )

(defun print-hier-node (hier node &optional (indent 0))
  (depth-indent indent)
  (print-it node " [" (get-id hier node) "," (get-max hier node) "] ")
  (dolist (x (get-sub hier node))
	  (print-hier-node hier x (+ indent 1)))
  )

(defun print-part-hier-node (hier node bounds &optional (indent 0)
				  &aux (min (car bounds))
				  (max (cadr bounds))
				  (total (cdr (assoc hier (getp node 'total))))
				  (mint (car total))
				  (maxt (cadr total)))
  (depth-indent indent)
  (print-it node " [" (get-id hier node) "," (get-max hier node) "] " #\tab)
  (cond ((equal min max) 
	 (if (equal min 1) 
	     (print-it "Exactly 1") 
	     (print-it "A set of exactly " min)))
	((and (equal min 0) (equal max 1)) (print-it "Optional"))
	((null max) (print-it "A set of at least " min))
	(t (print-it "A set of at least " min " and at most " max)))
  (print-it ", for a total ")
  (cond ((equal mint maxt) (print-it "of exactly " mint)) 
	((null maxt) (print-it "of at least " mint))
	(t (print-it "between " mint " and " maxt)))
  (dolist (x (cdr (assoc hier (getp node 'BOUNDS)))) 
	  (print-part-hier-node hier (car x) (cdr x) (+ indent 1)))
  )


; The next two routines (add-indicate and add-topic) add indicator links to
; predicates. Add-indicate adds a number of different topic links to a
; single predicate; add-topic adds topic links for a single topic to a number 
; of different predicates.
; Both have the same effect, but it may be more natural to use one or the
; other.


; add-indicate adds indicator links to a predicate
; arguments are
;    predicate indicator-list
; where items in indicator-list are either a single topic, or a list
; consisting of a topic and an argument number.
; eg. (add-indicate 'kill '(tp.violent-action 1) '(tp.existance 2))
;     (add-indicate 'grey 'tp.coloring)

(defun add-indicate (&rest args &aux
			   (pred (car args))
			   (indicators (cdr args))
			   (pred-int 1)
			   int usei)
  (rewind-question)
  (dolist (i indicators)
	  ; indicator is either of form topic, or (topic #)
	  ; make sure they are really topics
	  (if (and (setq int (getp (if (listp i) (car i) i) 'pred-interest))
		  (> int pred-int))
	      (setq pred-int int)) 
	  (when (member 'TP.TOPICS (getp (if (listp i) (car i) i) 'HIER))
		(pushnew i usei :test #'equal)))
  (if (or (null (getp pred 'pred-interest))
	  (< (getp pred 'pred-interest) pred-int))
      (change-property pred 'pred-interest pred-int))
  (add-is pred usei)
  )

(defun add-is (item topics &aux 
			(oldi (getp item 'indicators)))
  (change-property item 'INDICATORS (union topics oldi :test 'equal))
  (dolist (s (get-all-sub item))
	  (add-is s topics))
  )

(defun add-i-list (item topic &aux 
			(oldi (getp item 'indicators)))
  (change-property item 'INDICATORS (pushnew topic oldi :test 'equal))
  (dolist (s (get-all-sub item))
	  (add-i-list s topic))
  )

; add-topic adds indicator links to a predicate
; arguments are
;    topic predicate-list
; where items in predicate-list are either predicates or argument numbers
; (a single number may follow a predicate in the list)a
; eg. (add-topic 'tp.location 'here 'in 1)
;     (add-topic 'tp.coloring 'grey 'red 'blue)

(defun add-topic (&rest args &aux
			(topic (car args))
			(preds (cdr args))
			(int (getp topic 'pred-interest))
			lastpred)
  (rewind-question)
  (if (null int) (setq int 1))
  (unless (member 'TP.TOPICS (getp topic 'HIER))
	  (print-line " " topic " is not on the TP.TOPICS hierarchy "
			 "- topic indicators not added")
	  (return-from add-topic nil))
  (dolist (p preds)
	  ; either pred or #
	  ; if pred, wait to see if next is number, let prev one go
	  (when (and (not (or (member p '(subject object iobj)) (numberp p))) 
		     (or (null (getp p 'pred-interest))
			 (> int (getp p 'pred-interest))))
		(change-property p 'pred-interest int))
	  (if (or (member p '(subject object iobj)) (numberp p))
	      ; put this topic-argument number pair on lastpred
	      (if lastpred
		  (let ()
		       (add-i-list lastpred (list topic p))
		       (setq lastpred nil))
		  (print-line "Need predicate before argument indicators"))
	      (let ()
		   ; put this topic on lastpred - no argument numbers
		   (when lastpred
			 (add-i-list lastpred topic)
			 ) 
		   (setq lastpred p))
	      )
	  )
  ; make sure they are all done
  (when lastpred
	(add-i-list lastpred topic)
	)
  )

; print out predicate information
(defun print-pred (brief pred)
  (declare (ignore brief))
  (print-line pred)
  (when (getp pred 'INDICATORS)
        (print-line "  Indicates: ")
	(dolist (i (getp pred 'INDICATORS)) (print-it i " "))
	)
  (when (getp pred 'HIER)
	(print-line "  Belongs to hierarchies: " (getp pred 'HIER))
	)
  (print-line "  Interest as predicate: " (getp pred 'pred-interest))
  (when (getp pred 'interest)
	(print-line "  Interest as subject: " (getp pred 'interest))
	)
  )

(set-display-function 'pred 'print-pred 
  "Displays topic indicators, hierarchies and interest for a predicate")

;hier-rel-in-same-hier returns the relationship between two nodes on the
; same hierarchy using the preorder number scheme.

(defun hier-rel-in-same-hier (hier a b &aux
				   id1 id2 max1 max2)
  (if (or (getp hier 'hier-change)
	  (null (get-id hier a)) (null (get-id hier b)))
      (number-hier hier))
  (setq id1 (get-id hier a))
  (setq id2 (get-id hier b))
  (setq max1 (get-max hier a))
  (setq max2 (get-max hier b))
	
  (cond ((or (> id1 max2) (< max1 id2))
	 (if (equal (getp hier 'HIER-TYPE) 'EXCLUSION)
	     'DISJOINT
	     'UNKNOWN))
	((= id1 id2) 
	 'EQUIVALENT)
	((> id1 id2) 
	 'SUBSUMED)
	((<= id1 id2)
	 'SUBSUMES)
	(t 'UNKNOWN))
  )

(defmacro hier-type-num (hier)
  `(if (equal (getp ,hier 'HIER-TYPE) 'EXCLUSION)
       0
       5)
  )

; sort-hiers sorts the hierarchies so that exclusion hiearchies come first
; because they are more likely to give an answer

(defun sort-hiers (hierlist)
  (sort (copy-list hierlist) '< :key #'(lambda (a) (hier-type-num a)))
  )

; returns a sorted list of connections - immeidate connections first, sorted
; with exlcusion hierarchies first, followed by other connections, also
; sorted by exlusion vs overlap. If the effort level isn't high enough,
; only the immediate connections will be returned.

(defun sort-conn (conn-list good-hiers &aux use-list alt-list)
  (dolist (c conn-list)
	  (if (member (car c) good-hiers :test 'equal)
	      (push c use-list)
	      (if (> *specialist-eval-effort* 1)
		  (push c alt-list)
		  )))
  (setq use-list (sort use-list '< :key #'(lambda (a)
     (hier-type-num (car a)))))
  (setq alt-list (sort alt-list '< :key #'(lambda (a)
     (hier-type-num (car a)))))
  (append use-list alt-list)
  )


; hier-rel determines the relationship between two items based on their 
; locations in a hierarchy.
; Both must belong to the same hierarchy to be compared
; For overlap hierarchies, disjointedness cannot be determine this way 
; -inference is required

(defun hier-rel (a b &optional allowed-parts &aux
		   (aparts (getp a 'pred-parts))
		   (bparts (getp b 'pred-parts))
		   a1 b1 a2 b2
		   (res 'UNKNOWN))
  (if (member (last1 aparts) *remainder-indicators*)
      (setq aparts nil))
  (if (member (last1 bparts) *remainder-indicators*)
      (setq bparts nil))
  (setq *allowed-part-hiers* allowed-parts)

  (setq res (hier-test a b (getp a 'hier) (getp b 'hier)))
  (when (and (or (null res) (equal res 'unknown))
	     (or aparts bparts))
	(cond (aparts (setq a1 (car aparts))
		      (setq a2 (cadr aparts)))
	      (t (setq a2 a)))
	(cond (bparts (setq b1 (car bparts))
		      (setq b2 (cadr bparts)))
	      (t (setq b2 b)))
	(cond ((and a1 b1)
	       (setq res (compare-hier-res (hier-rel a1 b1 allowed-parts)
					   (hier-rel a2 b2 allowed-parts))))
	      ((null b1)
	       (setq res (compare-hier-res 'subsumed
					   (hier-rel a2 b2 allowed-parts))))
	      (t (setq res (compare-hier-res 'subsumes
					     (hier-rel a2 b2 allowed-parts)))))
	)

  ; also try projections if effort high enough and we don't have answer yet
  ; require exhaustive breakdowns then, or similar mechanism

  (setq *allowed-part-hiers* nil)
  res)

(defun hier-test (a b ahier bhier &aux (res 'unknown))
  ; first see if both have a common hierarchy - if so, try that
  (if (equal a 'entity)
      (dolist (h bhier) (if (getp h 'hier-change) (number-hier h))))
  (if (equal b 'entity)
      (dolist (h ahier) (if (getp h 'hier-change) (number-hier h))))
  (dolist (h (sort-hiers (remove-if #'(lambda (a) 
		(and (getp a 'part-hier) (not (member a *allowed-part-hiers*))))
				    (intersection ahier bhier))))
	  (setq res (hier-rel-in-same-hier h a b))
	  (unless (equal res 'UNKNOWN) (return))
	  )

  ; If that didn't give an answer, try using "connections" between hierarchies
  ; depending on effort level
  (when (and (equal res 'UNKNOWN) (> *specialist-eval-effort* 0))
	(dolist (ah (sort-hiers ahier))
		(unless (and (getp ah 'part-hier) 
			     (not (member ah *allowed-part-hiers*)))
		(setq res (find-hier-path a ah b 
					  (> *specialist-eval-effort* 1)))
		(unless (equal res 'UNKNOWN) (return res))) )
	; if no path can be found from a to b, is there likely to be 
	; apath in reverse? not the way things are done here I think
	; path can be found subsumes, subsumed, disjoint from either way
	)
  res)

(defun reverse-hier-rel (rel)
  (case rel
	((UNKNOWN DISJOINT) rel)
	(SUBSUMES 'SUBSUMED)
	(SUBSUMED 'SUBSUMES)
	(otherwise 'UNKNOWN))
  )

(defun compare-hier-res (resa resb)
  (cond ((or (equal resa 'UNKNOWN) (equal resb 'UNKNOWN))
	 'UNKNOWN)
	((and (equal resa 'DISJOINT) (equal resb 'DISJOINT))
	 'UNKNOWN)
	((equal resa resb) resa)  ;subsumes, subsumed, equivalent
	((equal resa 'EQUIVALENT) resb)
	((equal resb 'EQUIVALENT) resa)
	((or (and (equal resa 'DISJOINT) (equal resb 'SUBSUMES))
	     (and (equal resa 'SUBSUMED) (equal resb 'DISJOINT)))
	 'DISJOINT)
	(t 'UNKNOWN))
  )

; if "follow" is t, then we can look for a path involving more than one 
; intermediate node, otherwise we can only use one connection.
; If there is a path involving only one node from b to a, then there is
; one involving only one node from a to b, so we don't have to worry about
; the search in one direction going on too long when it would be faster
; in the other direction.

(defun find-hier-path (a hiera b follow &optional (sofar 'EQUIVALENT) 
			 (hiers-seen (list hiera))
			 &aux (conn-list (getp hiera 'CONNECTIONS))
			 (res 'UNKNOWN)
			 newhier resan newsofar)
  (dolist (c (sort-conn conn-list (getp b 'HIER)))
	  (setq newhier (car c))
	  (unless (or (member newhier hiers-seen)
		      (and (getp newhier 'part-hier)
			   (not (member newhier *allowed-part-hiers*))))
	  (dolist (n (cdr c))
		  (setq resan (hier-rel-in-same-hier hiera a n))
		  (setq newsofar (compare-hier-res sofar resan))
		  (unless (equal newsofar 'UNKNOWN)
			  (if (member newhier (getp b 'HIER))
			      (setq res (compare-hier-res newsofar
					  (hier-rel-in-same-hier newhier n b)))
			      ; try for rest of path
			      (if follow
				  (setq res (find-hier-path n newhier b follow
							    newsofar
					   (append hiers-seen (list newhier))))
				  (setq res 'UNKNOWN))
			      ))
		  (unless (equal res 'UNKNOWN) (return res))
		  ))
	  (unless (equal res 'UNKNOWN) (return res))
	  )
  res)

(defun add-sort-to-hier (item sort)
  (unless (equal sort (getp item 'sort))
	  (change-property item 'SORT sort)
	  (dolist (s (get-all-sub item))
		  (add-sort-to-hier s sort))
  ))

(defun add-part-hier (&rest args &aux
                            (hier (car args))
                            (root (cadr args))
                            (subs (cddr args))
                            slist min max sub bound)
  ; if super of root on hier, this is not the root - must be a part. If it
  ; is the root, must be an individual entity
  (set-hier-type hier 'PART-HIER)
  (apply 'add-hier (mapcar #'(lambda (a) (if (listp a) (car a) a)) args))
  (change-property root 'PARTS (union (getp root 'PARTS) (list hier)))
  (dolist (s subs)
          (setq min nil)
          (setq max nil)
          (cond ((listp s)
                 (setq sub (car s))
                 (setq bound (cadr s))
                 (cond ((listp bound)
                        (setq min (car bound))
                        (setq max (cadr bound)))
                       (t (setq min (setq max bound))))
                 )
                (t (setq sub s)))
          (when (null min) (setq min (setq max 1)))
	  (when (and (remainder-part s) (equal min max) (equal min 1))
		(setq min 0) (setq max nil))
          (setq slist (combine-bounds slist sub min max))
          )
  (change-property root 'BOUNDS (add-bounds hier root slist))
  )

(defun remainder-part (item &aux pparts)
  (when (symbolp item)
	(setq pparts (getp item 'pred-parts))
	(and pparts (member (last1 pparts) *remainder-indicators*)))
  )
 
 
(defun combine-bounds (slist sub min max &aux
                             (newbound (list min max)))
  (if (assoc sub slist)
      (nondest-rplacd slist sub newbound)
      (acons sub newbound slist))
  )

(defun add-bounds (hier node slist &aux
                        (olds (getp node 'BOUNDS)))
  (if (assoc hier olds)
      (nondest-rplacd olds hier slist)
      (acons hier slist olds))
  )

(defun calc-part-totals (hier &aux 
			      (root (getp hier 'root)))
  (node-part-total hier root '(1 1))
  )

(defun node-part-total (hier node sofar &aux subs)
  (change-property node 'TOTAL (add-num hier node sofar 'TOTAL))
  (setq subs (cdr (assoc hier (getp node 'bounds))))
  (mapc #'(lambda (h) 
     (node-part-total hier (car h) 
		      (list (* (car sofar) (cadr h)) 
			    (if (and (cadr sofar) (caddr h)) 
				(* (cadr sofar) (caddr h)))) ))
	subs)
  )

(set-display-function 'familiar-parts 'familiar-print "Prints familiar parts")
(set-display-function 'familiar 'familiar-print "Prints familiar parts")

(defun familiar-print (brief &rest args &aux n)
  (declare (ignore brief))
  (if (null args) (setq args *familiar-parts*))
  (print-line "Familiar Part" #\tab "Normal Levels")
  (dolist (a args)
	  (print-line a #\tab #\tab) 
	  (cond ((setq n (getp a 'normal-level)) 
		 (print-it (caar n) " " (cdar n))
		 (dolist (nl (cdr n)) (print-it ",   " (car nl) " " (cdr nl) )))
		(t (print-it "no normal levels"))))
  )
