;       Last change    August 1993
(in-package epilib)

; Interest calculations and tests
; Wff interest is calculated based on predicates and arguments' current interest
; and then remains stable unless boosted through back-propagation through a
; causal chain. Individual interest gets updated by each additional wff 
; involving the individual so increases over time. 
; Type predicates may have two types of interest - as subject (for use when
; calculating variable interest), and as predicate (for normal predications)
; Wffs also have two types of interest - their own intrinsic interestingness
; and an inherited interest component from the parent used to generate them

(tweakable '*operator-component* 1 
  "Weighting given to predicate/operator/function during interest calculations"
  '(float * *) )
(tweakable '*arg-component* 1 
  "Weighting given to arguments during interest calculations" '(float * *))
(tweakable '*wff-component* 0.1 
  "How much of wff interest is used to update individual interest" 
  '(float * *))
(tweakable '*interest-threshold* 0
  "Number beyond which wffs are considered interesting" '(float * *))
(tweakable '*check-inherit* t
  "Allow forward inferencing to continue based on inherited values"
  'boolean)
(tweakable '*inherit-amount* 5
  "Amount to decrease inherited interest by at each step" '(float * *))
(tweakable '*qa-inherit-amount* 10
  "Amount to decrease inherited interest by at each step" '(float * *))
(tweakable '*initial-charge* 0
  "Extra amount to add for inherited interest of story input" '(float * *)) 
(tweakable '*back-up-interest* nil 
  "Tells whether or not to back up interest of caused wff to causing wff"
  'boolean)
(tweakable '*deep-thought* 1 
  "Low numbers (< 1) indicate shallow thinking, high numbers deep thinking"
  '(float * *))
(tweakable '*minimum-interest* 1 "Minimum interest allowed" '(float * *))
(tweakable '*maximum-interest* 1000 "Maximum interest allowed" '(float * *))


(defvar *constants-to-update* nil "Constants to update interest for")

;
; Initial calculations of interest 
;

(defun wff-interest (wff)
  (case (getp wff 'wff-type)
	(variable 0)
	(constant (interest wff))
	(quantified
	 ; take maximum of restriction and main?
	 (if (getp wff 'restrict)
	     (max (wff-interest (getp wff 'restrict))
		  (wff-interest (getp wff 'main)))
	     (wff-interest (getp wff 'main)) 
	     ))
	(modified (calc-interest (getp wff 'operator) (getp wff 'arguments)))
	(logical
	 ; conditional - take interest of conclusion?
	 ; or - take average of arguments?
	 ; and - take maximum of arguments?
	 (cond ((getp wff 'conditional)
		(wff-interest (cadr (getp wff 'arguments))))
	       ((equal (getp wff 'pred) 'or)
		(apply 'average (mapcar 'wff-interest (getp wff 'arguments))))
	       (t (apply 'max (mapcar 'wff-interest (getp wff 'arguments))))
	       ))
	(episodic 
	 (let* ((args (getp wff 'arguments))
		(int (wff-interest (car args))))
	      (unless (not (equal (getp (cadr args) 'term-type) 'constant))
		      (pushnew (cadr args) *constants-to-update* :test 'equal)
		      )
	       (if (and (equal (getp wff 'pred) '_) *qa-mode*)
		   (* -1 int)
		   int)) )
	((causal prefix) (calc-interest (getp wff 'pred) (getp wff 'arguments)))
	(otherwise (interest wff))
	)
  )

(defun term-interest (term &optional type-lit)
  (cond ((symbolp term)
	 (case (getp term 'term-type)
	       (variable (variable-interest term type-lit))
	       (constant 
		(pushnew term *constants-to-update* :test 'equal)
		(interest term 1))
	       ((function modified)
		(calc-interest (getp term 'operator) (getp term 'arguments)))
	       ((quoted-expression record quasi-quoted-expression) 0)
	       (wff (wff-interest term))
	       (otherwise (interest term 1))
	       )
	 )
	(t 0))
  )

(defun predicate-interest (pred)
  (cond ((getp pred 'pred-interest))
	(t 1))
  )

(defun pred-interest (pred)
  (cond ((symbolp pred)
	 (case (getp pred 'pred-type)
	       (variable 0) 
	       (constant (predicate-interest pred))
	       (modified
		(calc-interest (getp pred 'operator) (getp pred 'arguments)))
	       (lambda
		(if (getp pred 'quantifier)
		    (wff-interest pred)
		    (calc-interest (getp pred 'operator) 
				   (getp pred 'arguments)))
		)
	       (otherwise (predicate-interest pred))
	       ))
	((numberp pred) 1)
	(t 0))
  )

(defun op-interest (op)
  (cond ((symbolp op)
	 (case (getp op 'op-type)
	       (variable 0) 
	       (constant (predicate-interest op))
	       (modified
		(calc-interest (getp op 'operator) (getp op 'arguments)))
	       (otherwise (predicate-interest op))
	       ))
	(t 0))
  )

(defun item-interest (item &optional junk)
  (cond ((symbolp item)
	 (case (entity-type item)
	       (wff (wff-interest item))
	       (pred (pred-interest item))
	       (term (term-interest item junk))
	       (op (op-interest item))
	       (otherwise (term-interest item))
	       ))
	(t 0)
	)
  )

(defun calc-interest (operator args &aux
			       arg-weights arg-ints pred-int wt
			       (type-lit (type-pred-p operator)))
  ; multiply operator interest by weighted contribution 
  ; from each argument
  ; for complex operators/preds - put weights on as built? or use
  ; another method to get weights here
  (setq arg-weights (cond ((and (symbolp operator) 
				(setq wt (getp operator 'weights)))
			   wt)
			  (t (make-weight-list (list-length args)))))
  (setq arg-ints (mapcar #'(lambda (a) (item-interest a type-lit)) args))
  (setq pred-int (item-interest operator))
	       
  (* (* pred-int *operator-component*)
     (* (apply '+ (mapcar '* arg-ints arg-weights)) 
	*arg-component*)
     )
  )

(defun variable-interest (var &optional type-lit &aux i
			      (ts (append (getp var 'types)
					  (if (getp var 'sort)
					      (list (getp var 'sort))))))
  (cond ((getp var 'interest)) 
	(ts
	 (apply 'max (mapcar #'(lambda (a) (if type-lit 1
					 (if (> (setq i (interest a)) 0) i
					     1))) ts)))
	(t 1))
  )

; sets the weights to all be the same
(defun make-weight-list (length)
  (make-sequence 'list length :initial-element 
		 (float (/ 1 (if (> length 0) length 1))))
  )


; 
; Setting/changing/updating interest values of wffs, individuals and type
; predicates
;

(defun set-interest (wff &aux int c-int)
  (when (symbolp wff)
	(setq *constants-to-update* nil)

	; calculate and set interest
	(setq int (update-interest-num (wff-interest wff)))
	(setp wff 'interest int)
	(setp wff 'about *constants-to-update*)

	; propagate interest up to causal predecessors?
	; do this before updating constants - this may add some more to
	; the list and then we don't have to worry about updating anything
	; twice

	(when (and *back-up-interest* (getp wff 'input-index))
	      (back-up-interest wff int) )

	; update the interests of constants involved
	(setq c-int (* int *wff-component*)) 
	(dolist (c *constants-to-update*) (update-constant-interest c c-int)) 
	int)
  )

(defun update-interest-num (int)
  (cond ((and *minimum-interest* (< int *minimum-interest*)) *minimum-interest*)
	((and *maximum-interest* (> int *maximum-interest*)) *maximum-interest*)
	(t int))
  )

(defun update-constant-interest (constant int &aux
					  (oldi (getp constant 'interest))
					  new-int)
  (when (type-pred-p constant)
	(when (or (null oldi) (> int oldi))
	      (change-property constant 'interest int))
	(return-from update-constant-interest))

  ; add a portion of the wff interest to the permanent individual interest
  (setq new-int (update-interest-num (+ (if oldi oldi 0) int)))

  (change-property constant 'interest new-int)

  ; update interest of types "as subject"
  (dolist (tp (grab-types constant)) (update-type-interest tp new-int))

  new-int)

(defun update-type-interest (type int &aux
				  (oldi (getp type 'interest)))
  (when (or (null oldi) (> int oldi))
	(change-property type 'interest int)
	(dolist (s (get-all-super type)) (update-type-interest s int))
	)
  )

(defun interest (item &optional (default 0) &aux i)
  (cond ((type-pred-p item) (if (setq i (getp item 'pred-interest)) i 1))
	((getp item 'interest))
	(t default))
  )

; inherited interest does not change if there is a boost of interest within 
; a chain of forward inferences - not sure yet if it should or not
(defun inherit-int (item newprob &aux useint 
			 (p1 (getp item 'c-prob)))
  (setq useint (cond ((getp item 'inherit-int))
		     (t (setp item 'inherit-int
			      (+ (interest item) *initial-charge*))
			(getp item 'inherit-int))))
  ; now degrade the inherited interest by 1 for the step
  ; also degrade an amount for a decrease in probability
  ; (an corresponding change for probability increase is not done)

  (inherit-calc useint p1 newprob)
  )

(defun inherit-calc (useint oldprob newprob &aux df)
  ; now degrade the inherited interest by 1 for the step
  ; also degrade an amount for a decrease in probability
  ; (an corresponding change for probability increase is not done)

  (when (< newprob oldprob)
	(setq df (- (log newprob 2) (log oldprob 2)))
	(if (> df 0) (setq useint (- useint df)))
	)

  (if *qa-mode*
      (setq useint (- useint *qa-inherit-amount*))
      (setq useint (- useint *inherit-amount*)))
  (if (< useint 0) (setq useint 0))

  useint)


; 
; Interest comparisons
;
(defun check-interest (wff &aux
			   (int (interest wff)))

  ; two components to check 
  ;   is there enough inherited interest to carry this one through?
  ;   or is it high enough all by itself?
  (or (and *check-inherit*
	   (> (cond ((getp wff 'inherit-int))
	       ; if no inherited interest has been assigned this is an
	       ; input clause so use its own interest
	       (t int))
	      0))
      (> (* int (getp wff 'c-prob) *deep-thought*) *interest-threshold*)
      )
  )

;
; Initialization of interest values for predicates/topics
;

(defun add-interest (item interest &aux
			  (oldi (getp item 'pred-interest)))
  (when (or (not (integerp interest))
	    (< interest 1)
	    (> interest 10))
	(return-from add-interest (print-error "Interest " interest 
		" must be an integer between 1 and 10")) )
  (when (and oldi (>= oldi interest))            
	(return-from add-interest nil))

  (change-interest item interest)
  )

(defun change-interest (item interest)
  (change-property item 'pred-INTEREST interest)
  (dolist (s (get-all-sub item))
	  (when (or (null (getp s 'pred-INTEREST))
		    (< (getp s 'pred-INTEREST) interest))
		(change-interest s interest))
	  )
  interest
  )


;
; Backing interest up through causal chains
;

(defun last-class (cl)
  (cond ((atom (cdr cl)) cl)
	((null (cdr cl)) cl)
	(t (if (cddr cl)
	     (last-class (cdr cl))
	     (last-class (cadr cl))))
	)
  )

(defun back-up-interest (wff int &aux
			     cint 
			     (cause (cause-of-wff wff))
			     (about (getp wff 'about)))
  (dolist (a1 about)
	  (dolist (cwff (get-class (if cause
				       (list a1 'tp.description)
				     (list a1 'tp.causal-relationship))
				   t))
		  (setq cint (getp cwff 'interest))
		  ; check direction
		  (when (and (or (null cint) (> int cint)) (traced-p 'forward)
			     (cause-test-ok a1 cwff cause))
		      (print-line "Backing up interest " int " to "
				     cwff ": " (getp cwff 'print))
		      (setp cwff 'interest int)
		      (setq *constants-to-update*
			    (union *constants-to-update* (getp cwff 'about)
				   :test 'equal))
		      (back-up-interest cwff int)
		      ))
	  )
  )

(defun cause-test-ok (a1 wff cause)
  ; make sure that wff is a cause wff, and if so, that a1 is on the proper
  ; end of it
  (if (or cause (cause-of-wff wff))
      (equal a1 (cadr (getp wff 'arguments)))
      )
  )


(defun cause-of-wff (wff &aux wt pred args a1)
  (and (symbolp wff) 
       (cond ((member (setq wt (getp wff 'wff-type)) '(prefix causal))
	      (or (member (setq pred (getp wff 'pred)) 
			  '(because cause-of involve))
		  (and (equal (getp pred 'pred-type) 'lambda)
		       (getp pred 'wff-type)
		       (cause-of-wff pred))))
	     ((equal wt 'modified)
	      (and (equal (length (setq args (getp wff 'arguments))) 1)
		   (symbolp (setq a1 (car args)))
		   (equal (entity-type a1) 'wff)
		   (cause-of-wff a1)))
	     (t nil))
       )
  )

