;;; The EpiShell input-output package.  A miscellaneous collection of 
;;; io routines.
(in-package user)
(eval-when (compile load eval)   
(provide 'esh-io)
(defpackage esh-io 
  (:use lisp #+lucid lucid-common-lisp #-lucid common-lisp #+excl excl)
  (:export
          err-msg 		;print msgs on *error-output*, prepended by
				; the current file name	
	  case-filename 	;transform a filename according to 
				; *esh-filename-case*
	  print-trace	        ;print msgs on *esh-trace-output*
          print-msg		;print msgs on *standard-output* 
	  error-msg 	        ;print msgs on *error-output*
	  context-msg     	;print msgs prepended by given file name
	  context-err-msg	;print msg son *error-ouput*, prepended by 
				; given file name
          file-open-msg 	;standard esh file open error message
	  get-filename 	        ;get the filename associated with an object
	  check-filename 	;check a file object for validity, and against
				; a given list of currently open files
	  open-file		;open a file
	  esh-eof 		;standard esh end-of-file constant
	  esh-read-error	;standard esh read-error constant

	  *esh-trace-output* *esh-filename-case* *esh-error-prefix*


	 ))
(unless (find-package 'esh-io) (make-package 'esh-io)))
(in-package esh-io)


(defconstant esh-eof '|esh eof| 
	"Returned by esh read functions to signal end-of file" )

(defconstant esh-read-error '|esh read eror| 
	"Returned by esh read functions to signal an error" )


(defvar *esh-trace-output* *standard-output*
  "Output stream for trace information" )

(defvar *esh-error-prefix* "esh: "
  "Prefix for EpiShell error messages" )

(defvar *esh-filename-case* :downcase
  "Case for filenames" )   ; if not :downcase or :upcase, no case
                           ; conversion is done on filenames

(defun err-msg (&rest msgs)
    (if *esh-error-prefix* 
        (princ *esh-error-prefix*))
    (apply #'context-err-msg (cons *standard-input* msgs))
)

(defun case-filename (filename)
    (case *esh-filename-case*
	(:downcase (string-downcase (string filename)))
	(:upcase   (string-upcase (string filename)))
	(otherwise (string filename)))
)

(defun print-msg (&rest msgs)
    (mapc #'(lambda (msg) (if msg (princ msg) (terpri)))
	  msgs )
    (fresh-line)
    nil
)

(defun print-trace (&rest msgs)
    (mapc #'(lambda (msg) (if msg (princ msg *esh-trace-output*) 
				  (terpri *esh-trace-output*)))
	  msgs )
    (fresh-line *esh-trace-output*)
    nil
)

(defun error-msg (&rest msgs)
    (mapc #'(lambda (msg) (if msg (princ msg *error-output*)
				  (terpri *error-output*)))
	  msgs )
    (fresh-line *error-output*)
    nil
)

(defun context-msg (file-context &rest msgs)
    (when (setq file-context (get-filename file-context))
	  (princ (file-namestring file-context)) (princ ": "))
    (apply #'print-msg msgs)
)

(defun context-err-msg (file-context &rest msgs)
    (when (setq file-context (get-filename file-context))
	  (princ (file-namestring file-context) *error-output*)
	  (princ ": " *error-output*))
    (apply #'error-msg msgs)
)

(defun file-open-msg (file-context file)
    (context-err-msg file-context "Could not open "
			              (or (get-filename file) "file.") )
)

(defun get-filename (file)
    (cond ((listp file)
		nil)
                          ; probe-file now demands that it argument be
                          ; a FILE stream. The only predicate I can find
                          ; to check this is a SYSTEM DEPENDENT one. 
;	  ((or (pathnamep file) (and (streamp file) (probe-file file))) 
	  ((or (pathnamep file) (and #+lucid (system:file-stream-p file)
				     #+allegro nil ;(and (streamp file) (not (interactive-stream-p file)))    
                                     (probe-file file)))
		(remove #\" (namestring file )))
	  ((or (stringp file) (symbolp file)) 
		(remove #\" (namestring (merge-pathnames 
				(case-filename (princ-to-string file))))))
	  (t nil))
)

(defun check-filename (given-name open-files &optional (file-context nil)
					   &aux filename)
   (if (not (setq filename (get-filename given-name)))
	(context-msg file-context given-name " is an invalid filename.")
	(if (member filename open-files :test #'equal)
	    (context-msg file-context filename " is already open.")
	    filename ) )
)

(defun open-file (filename access-type)
    (if filename 
         (apply #'open (cons filename
                    (cond ((eq access-type :input)
				'(:direction :input :if-does-not-exist nil))
                          ((eq access-type :output)
                                '(:direction :output :if-exists :new-version))
			  (t 
				'(:direction :output :if-exists :append
					:if-does-not-exist :create ))))))
)


