;       Last change    August 1993
(in-package lib)

;; Tweak functions set up user-tweakable variables with descriptions 

;; Useful functions
;;     tweak item newvalue
;;       changes the value of item to newvalue
;;       if newvalue is omitted, item is set to its default
;;     print-tweakables list
;;       prints the descriptions, defaults and current values for each
;;       tweakable item in list
;;       if list is ommitted, all tweakable items are printed

;; Useful global variables and constants:
;;     *tweakable-items*
;;        the list of tweakable items


(defvar *tweakable-items* nil "List of items whose value user can change")

; (deftype boolean () '(member t nil))  ; commented out [2/2/95 CH]

(defun all-symbols (lst)
    (reduce #'(lambda (a b) (and a (typep b 'symbol))) lst
            :initial-value t)) 
(deftype symbol-list () '(and list (satisfies all-symbols)))

(defun all-characters (lst)
    (reduce #'(lambda (a b) (and a (typep b 'character))) lst
            :initial-value t)) 
(deftype character-list () '(and list (satisfies all-characters)))

; tweakable adds a new item to the list of user-tweakable items
; It creates a variable with that name, and gives it th default value.
; The default and description are saved for later use.
; A LISP type specification can also be added to constrain possible
; values for the parameter.  Examples:  
;              INTEGER            - simple integer
;              (INTEGER 1 100)    - integer with range
;              BOOLEAN            - t or nil
;              FLOAT              - simple float
;              (FLOAT -10.0 10.0) - float with range
;              SYMBOL             - symbol
;              SYMBOL-LIST        - a list of symbols
;              CHARACTER-LIST     - a list of characters
;
; The function TWEAK will use this specification when assigning new
; values to the parameter.  TWEAK will coerce INTEGER values to
; FLOAT if necessary.
;
(defun tweakable (item &optional default description (type-spec nil))
  (pushnew item *tweakable-items*)
  (unless (boundp item)
      (eval (list 'defvar item)) )
  (when description (setf (documentation item 'setf) description))
  (setp item 'TYPE-SPEC type-spec)
  (setp item 'DEFAULT default)
  (tweak item)  ; make the default the initial value (and type-check it too)
  )

; tweak changes the current value of a tweakable item
(defun tweak (&optional item (value 'default) &aux type-spec) 
  (cond ((null item) (print-tweakables))

	((member item *tweakable-items*)
             ; 'default is the default tester, since nil could be a perfectly
             ; legitimate value for a tweakable item
	     (if (equal value 'default)
 		 (setq value (getp item 'DEFAULT)))
             (if (or (null (setq type-spec (get item 'TYPE-SPEC)))
                     (or (typep value type-spec)
                         (and (typep value 'integer) 
                              (setq value (coerce value 'float))
                              (typep value type-spec))))
                 (eval (list 'setq item (list 'quote value)))
                 (print-error "Values for item " item " must be of type "
                                  type-spec)))
	 (t (print-error item " is not a tweakable item. "
			  *nl* " Use (display '-k 'tweak) or check in the user manual.")))
  )


; print-tweakables prints all tweakable items or a selection of them
; given by the user. The item's name, its description, and its default and
; current values are printed.

(defun print-tweakables (&rest items)
  (if items 
      (setq items (intersection items *tweakable-items*))
      (setq items *tweakable-items*))
  (dolist (item items)
	  (print-it *nl* " "  (symbol-name item) ": " (documentation item 'setf)
                    *nl* "   Type: " (cond ((getp item 'TYPE-SPEC))
                                           (t "not specified"))
		    *nl* "   Default value: " (getp item 'DEFAULT)
		    *nl* "   Current value: " (eval item))
	  )
  )

(defun pr-tweak (brief &rest items)
  (cond ((not brief) (apply 'print-tweakables items))
	(t (if items 
	       (setq items (intersection items *tweakable-items*))
	       (setq items *tweakable-items*))
	   (dolist (item items)
		   (print-it *nl* " "  (symbol-name item) ": " 
			     (documentation item 'setf) ))
	   )) )

; these were originally defvar'd in the io file, but could not be made
; tweakable there without giving interdependencies
(tweakable '*print* t "Indicates whether or not to print anything" 'boolean)
(tweakable '*warn* t "Indicates whether or not to print warnings" 'boolean)

