;       Last change    August 1993
(in-package number)

; Evaluation routines

; find-path searches for a path from named number num1 to num2. It first checks
; paths stemming from strict links, and stops if it finds one, then from 
; non-strict.
; It returns nil if no path was found, less-than for a strict path, or lt= for
; a non-strict path

(defun find-path (net num1 num2 sofar already &aux
		       (slist (epinumber-strict-next num1))
		       (nslist (epinumber-next num1))
		       (name1 (epinumber-name num1))
		       (name2 (epinumber-name num2))
		       res lnum)
  (if (traced-p 'epilog::number-search)
      (print-line "    Number Specialist: Searching for path from "
		       name1 " to " name2))
  ; first try the strict links
  (dolist (l slist)
	  (setq lnum (findnumber net l))
	  (unless (member (epinumber-name lnum) already)
		  (if (traced-p 'epilog::number-search)
		      (print-line  "    Number Specialist: Using " 
				       "strict link between "
			               name1 " and " 
				       (epinumber-name lnum)))
	          (if (equal (epinumber-name lnum) name2)
		      ; found strict path - stop
		      (setq res 'less-than)
		      (setq res (find-path net lnum num2 'less-than 
				 (append already (list (epinumber-name lnum)))))
		      )
		  (if (member res '(less-than lt))
		      (return res))
		  )
	  )
  ; then try the non-strict links
  (unless (equal res 'less-than)
  (dolist (l nslist)
	  (setq lnum (findnumber net l))
	  (unless (member (epinumber-name lnum) already)
		  (if (traced-p 'epilog::number-search)
		      (print-line "    Number Specialist: Using " 
				       "non-strict link between "
			               name1 " and " 
				       (epinumber-name lnum)))
	          (if (equal (epinumber-name lnum) name2)
		      (setq res sofar)
		      (let ((res2 (find-path net lnum num2 sofar
				 (append already (list (epinumber-name lnum))))))
			   (unless (null res2)
				   (setq res res2))
			   )
		      )
		  (if (member res '(less-than lt))
		      (return res))
		  )
	  ))
  res
  )

; compare-reln determines the relation (equivalent, subsumes, etc) between
; the given relations

(defun compare-reln (reln1 reln2)
  (cond ((equal reln2 'unknown)
	 'unknown)
	((equal reln1 reln2)
         t)
        (t (case reln1 
	         (equal (if (member reln2 '(lt= gt=))
		            'unknown
		            nil))
	         ((less-than lt) (if (equal reln2 'lt=)
				     'unknown
				     (if (member reln2 '(lt less-than))
					 t
				         nil)))
	         ((greater-than gt) (if (equal reln2 'gt=)
		                        'unknown
				     (if (member reln2 '(gt greater-than))
					 t
				         nil)))
	         (lt= (if (member reln2 '(equal less-than lt))
		          t
		          (if (equal reln2 'gt=)
		              'unknown
		              nil)))
	         (gt= (if (member reln2 '(equal greater-than gt))
		          t
		          (if (equal reln2 'lt=)
		              'unknown
		              nil)))
		 )
	   )
	)
  )


; min-max-reln tries to determine the relaion between two named numbers by
; comparing their bounds.

(defun min-max-reln (num1 num2 &aux
			  (min1 (epinumber-min num1))
			  (min2 (epinumber-min num2))
			  (max1 (epinumber-max num1))
			  (max2 (epinumber-max num2))
			  (res 'unknown))
  (if (and min1 min2
	   (equal min1 min2)
	   max1 max2
	   (equal max1 max2)
	   (equal min1 max1))
      (setq res 'equal)
      )
  (when (equal res 'unknown)
        (if (and min1 max2)
            (if (> min1 max2)
	        (setq res 'greater-than)
	        (if (>= min1 max2)
	            (setq res 'gt=))
	        )
            )
	)
  (when (equal res 'unknown)
        (if (and max1 min2)
            (if (< max1 min2)
                (setq res 'less-than)
	        (if (<= max1 min2)
	            (setq res 'lt=))
	        )
	    )
	)
  (if (and (traced-p 'epilog::number-eval)
	   (not (equal res 'unknown)))
      (print-line "    Number Specialist: Found relation " 
		       (epinumber-name num1) " " res " " (epinumber-name num2)))
  res
  )

; find-reln tries to find the relation between tow named numbers by finding
; a path between them. If there isn't a path in one direction (num1 to num2)
; it tries the other way (num2 to num1). It returns the relation suggested
; by the path is one is found.

(defun find-reln (num1 num2 &aux
		       res)
  (setq res (find-path current-net num1 num2 'lt= (list (epinumber-name num1))))
  (if (or (null res)
	  (equal res 'unknown))
      (setq res (reverse-reln (find-path current-net num2 num1 'lt= 
					 (list (epinumber-name num2)))))
      res)
  (if (null res)
      (setq res 'unknown))
  (if (and (traced-p 'epilog::number-eval)
	   (not (equal res 'unknown)))
      (print-line "    Number Specialist: Found relation " 
		       (epinumber-name num1) " " res " " (epinumber-name num2)))
  res
  )

; find-relation first tries to compare bounds to find a relation between
; the two given named numbers. If that fails, it looks for paths.

(defun find-relation (net arg1 arg2 &optional (effort *specialist-eval-effort*) &aux
		      (num1 (findnumber net arg1))
		      (num2 (findnumber net arg2)))
  (cond ((or (null num1) (null num2))
         'unknown)
        ((equal (epinumber-name num1) (epinumber-name num2))
          'equal)
        ; try min-max comparisons
        (t (let ((cp (min-max-reln num1 num2)))
		(if (and (equal cp 'unknown)
		         (> effort 0))
                    ; try path search
	            (find-reln num1 num2)
	            cp
	            )
		))
	)
  )

; eval-reln determines if the given relation holds between the two arguments
; it returns t if it holds, nil if it doesn't, unknown otherwise

(defun eval-reln (net arg1 reln arg2 effort &aux
		      (a1 arg1)
		      (a2 arg2)
		      (num1 (findnumber net a1))
		      (num2 (findnumber net a2)))
  (cond ((or (null num1) (null num2))
         'unknown)
	((equal (epinumber-name num1) (epinumber-name num2))
	 (compare-reln reln 'equal))
	(t ; try min-max comparisons
	   (let ((cp (compare-reln reln (min-max-reln num1 num2))))
		(if (and (equal cp 'unknown)
		         (> effort 0))
                    ; try path search
	            (compare-reln reln (find-reln num1 num2))
	            cp
		    )
		)
	   )
      )
  )

; evaluate determines if the given literal is 	true, false or unknown
; it first ensures that the literal is in its domain

(defun evaluate (netname lit negp pred arglist effort &aux
		      (reln (find-symbol (symbol-name pred) 'number))
		      (net (determine-net netname))
		      (a1 (car arglist))
		      (a2 (cadr arglist))
		      (a3 (caddr arglist)))
  (declare (ignore lit))
  (if (and (null a3)
	   (not (non-constant-arg arglist))
           (number-arg a1)
	   (number-arg a2))
      (eval-with-neg net negp reln arglist effort)
      'unknown)
  )

; eval-with-neg evaluates the given literal, taking into account whether it
; is negated. If so, the positive literal is evaluated, and the result reversed. 
(defun eval-with-neg (net negp reln args effort &aux
			  (posans (evaluate-number net reln args effort)))
    
  (cond ((equal posans 'yes)
         (if negp
	     'no 
	     'yes))
	((equal posans 'no)
	 (if negp
	     'yes
	     'no))
	(t posans))
  )
 

; compare-min-max determines whether a relation holds between a named number
; and a number contsant

(defun compare-min-max (net a1 reln a2 &aux
			    (arg1 a1)
			    (arg2 a2)
			    (num (findnumber net arg1)))
  (if (null num)
      'unknown
      (compare-reln reln (nbr-reln num arg2))
      )
  )

; nbr-reln determines the relation between the named number num and the
; number constant val

(defun nbr-reln (num val &aux
			 (min (epinumber-min num))
			 (max (epinumber-max num)))
  (cond ((and min max (equal val min) (equal val max)) 'equal)
	((and max (> val max)) 'less-than)
	((and min (< val min)) 'greater-than)
	((and max (>= val max)) 'lt=)
	((and min (<= val min)) 'gt=)
	(t 'unknown)
	)
  )

; get-nbr-reln uses nbr-reln to check the relation between a named number
; and a number constant, after checking to make sure the number is in hte
; number graph

(defun get-nbr-reln (net a1 val &aux
			 (num (findnumber net a1)))
  (if (null num)
      'unknown
      (nbr-reln num val))
  )


; compare-numbers determines if the relation reln holds between the 
; number constants n1 and n2

(defun compare-numbers (n1 reln n2)
  (case reln 
	(equal (equal n1 n2))
	((less-than lt) (< n1 n2))
	((greater-than gt) (> n1 n2))
	(lt= (<= n1 n2))
	(gt= (>= n1 n2))
	)
  )

; cp-nbrs determines the relation between number constants n1 and n2

(defun cp-nbrs (n1 n2)
  (cond ((equal n1 n2) 'equal)
	((< n1 n2) 'less-than)
	((> n1 n2) 'greater-than)
	((<= n1 n2) 'lt=)
	((>= n1 n2) 'gt=)
	(t 'unknown))
  )

; evaluate-number determines whether the given relation holds betwen the
; arguments.
; If the arguments are both nubmer constants, compare-numbers is used;
; if one is a constant and the other a named number, compare-min-max is used,
; if both are named numbers, eval-reln is used

(defun evaluate-number (net reln arglist effort &aux
			    (a1 (car arglist))
			    (a2 (cadr arglist))
			    (res 'unknown))
  (unless (null arglist)
	  (when (traced-p 'epilog::number-search)
	        (print-line "    Number specialist: Evaluating " 
	  		       (print-info a1) " " reln " " (print-info a2))
		(if (not (equal (number-subnet-identifier net) '/))   
		    (print-it "    in subnet " 
			      (number-subnet-identifier net)))
		)
	  (setq res
	        (if (symbolp a1)
	            (if (symbolp a2)
		        (eval-reln net a1 reln a2 effort)
		        (compare-min-max net a1 reln a2))
	            (if (symbolp a2)
		        (compare-min-max net a2 (reverse-reln reln) a1)
		        (compare-numbers a1 reln a2))  
		    )
		)
	)
  ; set answers to yes, no or unknown
  (if (equal res t)
      (setq res 'yes))
  (if (null res)
      (setq res 'no))
  (if (and (traced-p 'epilog::number-eval)
	   (not (equal res 'unknown)))
      (print-line "    Number Specialist: Evaluated " 
	  		       a1 " " reln " " a2 " to " res))
  res
  )
