;       Last change    August 1993
(in-package time)

;            ABSOLUTE TIMES and DURATIONS

; Absolute times are used to determine time relationships in constant time
; when the psuedo time comparsions don't give an answer. When entered,
; absolute time minimums are propagated forward to all descendants since
; any point after the given one cannot possibly have an absolute time earlier.
; Absolute time maximums are propagated backward since any point before this
; one accnot have an absolute time greater than it.

; strict and nonstrict have no meaning here, as we are dealing
; in seconds and microseconds, and natural language speakers rarely
; get that precise

; new absolute times do not update durations. However durations can
; update absolute times

; for t1 t2 lower l1 l2, upper u1 u2
; t2 - t1 lower l, upper u
; l2 - u <= t1 <= u2 - l      (use best of l2 -u and l1, u2- l and u1)
; l + l1 <= t2 <= u + u1      (use best of l + l1 and l2, u + u1 and u2)
; l2-u1 <= t2 - t1 <= u2 - l1 (use best of l2 - u1 and l, u2 - l1 and u)



; time-get-max returns the maximum of abs1 and abs2
; It compares year, then month, etc, until one is greater than the other
; or it is impossible to compare them

(defun time-get-max (abs1 abs2 &aux
		     (e1 (car abs1))
		     (e2 (car abs2)))
  (if (endp abs1)
      nil
      (if (numberp e1)
	  (if (numberp e2)
	      (if (> e1 e2)
		  abs1
		  (if (< e1 e2)
		      abs2
		      (append (list e1) (time-get-max (cdr abs1) (cdr abs2)))
		      )
		  )
	      abs1
	      )
	  (if (numberp e2)
	      abs2
	      (if (equal e1 e2)
		  (append (list e1) (time-get-max (cdr abs1) (cdr abs2)))
		  abs1)
	      )
	  )
      )
  )

(defun replace-unknowns (abs1 abs2) 
  (mapcar #'(lambda (a b) 
		    (if (numberp a)
			a
			(if (numberp b)
			    b
			    a)))
	  abs1 abs2)
  )

; choose-max returns the maximum of oldabs and newabs, as long as it is
; no greater than max (this is used to ensure that absolute mins are not > max)

(defun choose-max (oldabs newabs max)
  (if (null oldabs)
      newabs
      (if (null newabs)
	  oldabs
	  (let ((setabs (time-get-max oldabs newabs)))
               (if (not (equal (test-point-answer 'after 
					  (compare-abs setabs max)) 'yes))
                   setabs
                   oldabs)
	       )
	  )
      )
  )


; time-get-min returns the minimum of abs1 and abs2
; It compares year, then month, etc, until one is greater than the other
; or it is impossible to compare them

(defun time-get-min (abs1 abs2 &aux
		     (e1 (car abs1))
		     (e2 (car abs2)))
  (if (endp abs1)
      nil
      (if (numberp e1)
	  (if (numberp e2)
	      (if (< e1 e2)
		  abs1
		  (if (> e1 e2)
		      abs2
		      (append (list e1) (time-get-min (cdr abs1) (cdr abs2)))
		      )
		  )
	      abs1
	      )
	  (if (numberp e2)
	      abs2
	      (if (equal e1 e2)
		  (append (list e1) (time-get-min (cdr abs1) (cdr abs2)))
		  abs1)
	      )
	  )
      )
  )

; choose-min returns the minimum of oldabs and newabs, as long as it is
; no less than min (this is used to ensure that absolute maxs are not < min)

(defun choose-min (oldabs newabs min)
  (if (null oldabs)
      newabs
      (if (null newabs)
	  oldabs
	  (let ((setabs (time-get-min oldabs newabs)))
               (if (not (equal (test-point-answer 'before 
					  (compare-abs setabs min)) 'yes))
                   setabs
                   oldabs)
	       )
	  )
      )
  )

; Merge-absolute-min merges a new absolute time with the old - the maximum
; absolute time is taken if one is strictly greater than the other, and 
; constant or variable names are replaced by numbers if possible.

(defun merge-absolute-min (oldabs newabs max)
  (if (null oldabs)
      newabs
      (if (null newabs)
	  oldabs
	  (let ((newmin (choose-max (replace-unknowns oldabs newabs) newabs max)))
	       (if (equal newmin newabs)
		   newmin
		   oldabs))
	  )
      )
  )

; Merge-absolute-max merges a new absolute time with the old - the minimum
; absolute time is taken if one is strictly less than the other, and 
; constant or variable names are replaced by numbers if possible.

(defun merge-absolute-max (oldabs newabs min)
  (if (null oldabs)
      newabs
      (if (null newabs)
	  oldabs
	  (let ((newmax (choose-min (replace-unknowns oldabs newabs) newabs min)))
	       (if (equal newmax newabs)
		   newmax
		   oldabs))
	  )
      )
  )

; Compare-abs-elements compares the first element in two absolute times
; If one is greater than the other, the comparison is done. If both are
; equal, the comparison continues with the next element. If one is not a
; number, the comparison stops with unknown.

(defun compare-abs-elements (abs1 abs2 &aux
				  (a1 (car abs1))
				  (a2 (car abs2)))
  (if (endp abs1)
      'same-time ;'equal
      (if (equal a1 a2)
          (compare-abs-elements (cdr abs1) (cdr abs2))
          (if (numberp a1)
              (if (numberp a2)
	          (if (> a1 a2)
	              'after-1
		      'before-1)
	          'unknown)
              'unknown)
          )
      )
  )

; Compare-abs returns the relation between the two absolute times abs1 and abs2.
; This can be equal, unknown, before, or after. 
; Be careful here - this may return equal when there is an arc indicating
; before or after.

(defun compare-abs (abs1 abs2)
  (if (or (null abs1)
	  (null abs2))
      'unknown
      (compare-abs-elements abs1 abs2)
      )
  )

;; Absolute time conversion and "arithmetic"

;; The code that was originally here was buggy.  It called the 32nd
;; day of the year February 0, and it handled leap years incorrectly.
;; I have ripped it all out and replaced it with calls to the Common
;; Lisp standard functions {encode,decode}-universal-time.  The zero
;; date for Common Lisp is 1/1/1900, and dates before that can't be
;; handled, but Epilog used to be able to deal with dates back to
;; 1/1/1.  Adjusting to the CL functions is not as simple as
;; adding/subtracting 1900 to the year, every 4th century year is a
;; leap year, and 1900 is not a multiple of 400.  So I add/subtract
;; 2000 instead, but that doesn't mean dates B.C. can be handled
;; properly. -ANK

;; abs-int calculates the number of seconds for the absolute time abs
(defun abs-int (abslist)
  (let ((year (+ (nth 0 abslist) 2000))
	(month (nth 1 abslist))
	(day (nth 2 abslist))
	(hr (nth 3 abslist))
	(min (nth 4 abslist))
	(sec (nth 5 abslist)))
    (encode-universal-time sec min hr day month year 0)))

;; cvt-abs calculates the absolute date from the number of seconds num
(defun cvt-abs (seconds)
  (multiple-value-bind (sec min hr day month year)
      (decode-universal-time seconds 0)
    (list (- year 2000) month day hr min sec)))


; Absolute time conversion and "arithmetic"

; add-abs-dur calculates a new absolute time dur seconds after abs

(defun add-abs-dur (abs dur)
  ; make new absolute time dur after abs
  (if (or (null dur)
	  (equal dur '0)
	  (null abs)
	  (find-if-not #'numberp abs))
      abs
      (let* ((numabs (+ (abs-int abs) dur))
	     (newabs (cvt-abs numabs)))
	    newabs
	    )
      )
  )

; add-abs-dur calculates a new absolute time dur seconds before abs

(defun sub-abs-dur (abs dur)
  ; make new absolute time dur before abs
  (if (or (null dur)
	  (equal dur '0)
	  (null abs)
	  (find-if-not #'numberp abs))
      abs
      (let* ((numabs (- (abs-int abs) dur))
	     (newabs (cvt-abs numabs)))
	    newabs
	    )
      )
  )

; re-calc-abs-min calculates the absolute time duration seconds after propabs
; and takes the maximum of that and the oldabs

(defun re-calc-abs-min (propabs pt2 duration &aux
				(oldabs (time-node-absolute-min pt2))
				(max (time-node-absolute-max pt2))
				newabs)
  ; take max of propabs + duration, and oldabs
  (setq newabs (add-abs-dur propabs duration))
  (choose-max oldabs newabs max)
  )

; re-calc-abs-max calculates the absolute time duration seconds before propabs
; and takes the minimum of that and the oldabs

(defun re-calc-abs-max (propabs pt2 duration &aux
				(oldabs (time-node-absolute-max pt2))
				(min (time-node-absolute-min pt2))
				newabs)
  ; take min of propabs - duration, and oldabs
  (setq newabs (sub-abs-dur propabs duration))
  (choose-min oldabs newabs min)
  )

; calc-duration-min calculates the minimum duration between two absolute times

(defun calc-duration-min (abs1 abs2)
  (if (or (null abs1)
	  (null abs2))
      nil
    (let ((diff (- (abs-int abs2) (abs-int abs1))))
      (if (< diff 0)
	  '0
	diff)
      )
    )
  )
 
; calc-duration-max calculates the maximum duration between two absolute times
; this works ok for all numeric abs times, but not for ones with symbols

(defun calc-duration-max (abs1 abs2)
  (if (or (null abs1)
	  (null abs2))
      nil
      (if (or (member-if-not 'numberp abs1)
	      (member-if-not 'numberp abs2))
	  nil
          (- (abs-int abs2) (abs-int abs1))
	  )
      )
  )
 
; duration-min determines the appropriate duration to use - the given one or
; one calculated from the absolute times. If both are valid, the maximum is
; duration-min determines the appropriate duration to use - the given one or
; one calculated from the absolute times. If both are valid, the maximum is
; chosen.

(defun duration-min (d1 d2)
  (if (or (null d1)
	  (equal d1 '-infinity))
      (if (or (null d2)
	      (equal d2 '-infinity))
          '0
          d2)
      (if (or (null d2)
	      (equal d2 '-infinity))
	  d1
	  (min-num d1 d2))
      )
  )


; prop-min-to-point propagates the minimum absolute time to the next
; descendant (the from point of link) 

(defun prop-min-to-point (link &aux
			       (pt1 (time-point
				     (time-link-from-point-name link)))
			       (pt2 (writeable-point
				     (time-link-to-point-name link)))
			       (pt1abs (time-node-absolute-min pt1))
			       (pt1max (time-node-absolute-max pt1))
			       (pt2abs (time-node-absolute-min pt2))
			       (durmin (time-link-duration-min link))
			       (durabs (calc-duration-min pt1max pt2abs))
			       (usedur (duration-min durmin durabs))
			       newabs)
  (setq newabs (re-calc-abs-min pt1abs pt2 usedur))
  (when (not (equal newabs pt2abs))
        (if (and (traced-p 'epilog::abs-time-entry)
      	   (not (null newabs)))
            (print-line
	     "    Time Specialist: Propagating absolute time minimum "
	      	       newabs " to " (time-link-to-point-name link))
            )
	(spec-changed-concept (time-link-to-point-name link))
	(change-field 'time-node-absolute-min pt2 newabs)
	(prop-absmin pt2)
	)
  )

; prop-max-to-point propagates the maximum absolute time to the previous
; ancestor (the to point of link) 

(defun prop-max-to-point (link oldabs &aux
				      (pt1 (time-point
					    (time-link-to-point-name link)))
				      (pt2 (writeable-point
					    (time-link-from-point-name link)))
				      (pt1abs (time-node-absolute-max pt1))
				      (pt2min (time-node-absolute-min pt2))
				      (pt2abs (time-node-absolute-max pt2))
				      (durmin (time-link-duration-min link))
				      (durabs (calc-duration-min oldabs pt2min))
				      (usedur (duration-min durmin durabs))
				      newabs)
  (setq newabs (re-calc-abs-max pt1abs pt2 usedur))
  (when (not (equal newabs pt2abs))
        (if (and (traced-p 'epilog::abs-time-entry)
                 (not (null newabs)))
            (print-line
	     "    Time Specialist: Propagating absolute time maximum "
		             newabs " to " (from-name link))
            )
	(spec-changed-concept (from-name link))
	(change-field 'time-node-absolute-max pt2 newabs)
	(prop-absmax pt2 pt2abs)
	)
  )

; prop-absmin propagates absolute time minimum from the given point to
; any descendants it has

(defun prop-absmin (point &aux
			  (dlist (time-node-descendants point))
			  (xdlist (time-node-xdescendants point))
			  (ditem (car dlist)))
  ; propagate only to first in chain descendant - since it is recursive
  ; it will get the rest of the chain anyway

  (unless (null ditem)
          (prop-min-to-point ditem))

  ; propagate to all x descendants that are in the current environment.

  (dolist (xitem xdlist)
    (if (writeable-p (time-point (time-link-to-point-name xitem)))
	(prop-min-to-point xitem))
    )
  )


; prop-absmax propagates absolute time maximum from the given point to
; any ancestors it has

(defun prop-absmax (point oldabs &aux
			  (alist (time-node-ancestors point))
			  (xalist (time-node-xancestors point))
			  (aitem (car (last alist))))
  ; propagate only to first in chain ancestor - since it is recursive
  ; it will get the rest of the chain anyway

  (if aitem
      (prop-max-to-point aitem oldabs))

  ; propagate to all x descendants that are in an environment to which
  ; the current environment is accessible (including the current one
  ; itself).

  (dolist (xitem xalist)
    (if (writeable-p (time-point (time-link-to-point-name xitem)))
	(prop-max-to-point xitem oldabs))
    )
  )

; update-absolute-min adds a new absolute time to pt

(defun update-absolute-min (pt abs &aux
			       (max (time-node-absolute-max pt))
			       (oldabs (time-node-absolute-min pt))
			       (newabs (merge-absolute-min oldabs abs max)))
  (assert (writeable-p pt))
  (unless (equal oldabs newabs)
          (if (and (traced-p 'epilog::abs-time-entry)
	           (not (null abs)))
              (print-line
	       "    Time Specialist: Adding absolute time minimum " 
		               abs " to " (time-node-name pt))
              )
          (change-field 'time-node-absolute-min pt newabs)
          (prop-absmin pt))
  )

(defun get-min-abs (abs &aux ab)
  (mapcar #'(lambda (a) (if (numberp a)
			    a
			    (let ()
				 (add-interested-party a 'time-specialist)
			         (cond ((numberp (setq ab (spec-eval-fn 
					       (make-fn-arg 'epilog::min-of a))))
					ab)
			               (t a))
				 )
			    ))
	  abs)
  )

(defun get-max-abs (abs &aux ab)
  (mapcar #'(lambda (a) (if (numberp a)
			    a
			    (let ()
				 (add-interested-party a 'time-specialist)
			         (cond ((numberp (setq ab (spec-eval-fn 
					       (make-fn-arg 'epilog::max-of a))))
					ab)
			               (t a))
				 )
			    ))
	  abs)
  )

(defun add-absolute-min (t1 abs &aux
			     (absnew (get-min-abs abs))
			     (pt (time-point t1)))
  (when (null pt)
	(setq pt (add-single t1)))
  (update-absolute-min (writeable-point pt) absnew)
  )

; update-absolute-max adds a new absolute time to pt

(defun update-absolute-max (pt abs &aux
			       (min (time-node-absolute-min pt))
			       (oldabs (time-node-absolute-max pt))
			       (newabs (merge-absolute-max oldabs abs min)))
  (assert (writeable-p pt))
  (unless (equal oldabs newabs)
          (if (and (traced-p 'epilog::abs-time-entry)
                   (not (null abs)))
              (print-line
	       "    Time Specialist: Adding absolute time maximum " 
		               abs " to " (time-node-name pt))
              )
          (change-field 'time-node-absolute-max pt newabs)
          (prop-absmax pt oldabs))
  )

(defun add-absolute-max (t1 abs &aux
			     (absnew (get-max-abs abs))
			     (pt (time-point t1)))
  (when (null pt)
	(setq pt (add-single t1)))
  (update-absolute-max (writeable-point pt) absnew)
  )

; find-link finds the link between pt1 and pt2. If none exists, it adds one

(defun find-link (pt1 pt2 &aux
		      (chain2 (time-node-chain pt2))
		      (psuedo2 (time-node-psuedo pt2))
		      dlist link)
  (if (on-same-chain pt1 pt2)
      (setq dlist (time-node-descendants pt1))
      (setq dlist (time-node-xdescendants pt1))
      )
  (setq link
        (dolist (item dlist)
      	        (when (and (equal (to-chain-number item)
				  (meta-node-chain-number chain2))
		           (equal (to-psuedo item) psuedo2))
	              (return item)
		      )
	        ))
  (if (null link)
      (add-link	(writeable-point pt1)
		(writeable-point pt2)
		'|1|)
      link
      )
  )

; update-duration-min adds a minimum duration to a link and propagates
; absolute time if necessary

(defun update-duration-min (link d)
  (when (or (and (> d 0)
		 (not (time-link-strict link)))
	    (or (null (time-link-duration-min link))
		(> d (time-link-duration-min link))))
    (setq link (writeable-link link))
    (let ((pt1 (time-point (time-link-from-point-name link)))
	  (pt2 (time-point (time-link-to-point-name link))))
           ;; pt1 and pt2 are guaranteed to be writeable
      (when (and (> d 0)
		 (not (time-link-strict link)))
	(change-field 'time-link-strictness link t)
	(if (on-same-chain pt1 pt2)
	    (add-strictness pt1 pt2)
	  )
	)
      (when (or (null (time-link-duration-min link))
		(> d (time-link-duration-min link)))
        (if (traced-p 'epilog::time-duration-entry)
            (print-line "    Time Specialist: Adding duration minimum " 
			d " between " (from-name link)
			" and " (time-link-to-point-name link))
	  )
        (change-field 'time-link-duration-min link d)
	(update-absolute-max pt1 
			     (sub-abs-dur (time-node-absolute-max pt2) d))
	(update-absolute-min pt2
			     (add-abs-dur (time-node-absolute-min pt1) d))
	)
      )
    )
  )

; new-duration-min adds a duration minimum between pt1 and pt2

(defun new-duration-min (pt1 pt2 d &aux 
			     link)
  (setq link (find-link pt1 pt2))
  (update-duration-min link d)
  )

(defun add-duration-min (t1 t2 d &aux
			     (pt1 (time-point t1))
			     (pt2 (time-point t2)))
  (new-duration-min pt1 pt2 d)
  )

; update-duration-max adds a maximum duration to a link 

(defun update-duration-max (link d)
  (when (or (null (time-link-duration-max link))
 	    (< d (time-link-duration-max link)))
    (if (traced-p 'epilog::time-duration-entry)
	(print-line "    Time Specialist: Adding duration maximum " 
		    d " between " (from-name link)
		    " and " (time-link-to-point-name link))
      )
    (change-field 'time-link-duration-max 
		  (writeable-link link)
		  d)
    )
  )

; new-duration-max adds a duration maximum between pt1 and pt2

(defun new-duration-max (pt1 pt2 d &aux
			     link)
  (setq link (find-link pt1 pt2))
  (update-duration-max link d)
  )

(defun add-duration-max (t1 t2 d &aux
			     (pt1 (time-point t1))
			     (pt2 (time-point t2)))
  (new-duration-max pt1 pt2 d)
  )

