;       Last change    August 1993
(in-package time)

; the following routines are used to determine whether two literals are 
; compatible or incompatible

; usable returns t if the arg given is an absolute time or exists in the given 
; time net; otherwise nil

(defun usable (arg)
  (if (absolute-time arg)
      t
      (not (missing arg))
      )
  )

; unifiable is used by the unify routine to determine if two constant arguments
; are "unifiable". They do not have to unify in the typical sense - it is
; enough if there is some relation between them. Since it can be time consuming
; to calculate this relation, a short cut is used whioch doesn't necessarily
; guarantee a relation - however, it will be tested during the entry/test
; phase anyway. Why do it twice.

(defun unifiable (arg1 arg2)
  (if (equal arg1 arg2)
      t
    (and (usable arg1) (usable arg2))))

; determine enter decides which of two literals should be entered. If one
; is negated, the other is entered, if both are negated, neither can be 
; entered.
; If there is a choice, and one is a between predicate, enter that one, as
; it is harder to test with

(defun determine-enter (neg-p1 pclass1 neg-p2)
  (if neg-p2
      (if neg-p1
	  nil
	  '1)
      (if neg-p1
	  '2
          (if (equal pclass1 2)
	      '1
              '2)
	  )
      )
  )

; classify-pred returns a number corresponding to the "class" of the 
; predicate. This just divides the predicates up so that similar ones are
; treated the same.

(defun classify-pred (pred)
  (cond ((member pred '(equal before before-1 before-0 after after-1 after-0
                        during during-0-0 during-1-1 during-0-1 during-1-0
			during-1 during-0 during--1 during--0
                        contains contains-0-0 contains-1-1 contains-0-1 
			contains-1-0 contains-1 contains-0 contains--1 
			contains--0
                        overlaps overlaps-0-0 overlaps-1-1 overlaps-0-1 
			overlaps-1-0 overlaps-1 overlaps-0 overlaps--1 
			overlaps--0
                        overlapped-by overlapped-by-0-0 overlapped-by-1-1 
			overlapped-by-0-1 overlapped-by-1-0 overlapped-by-1 
			overlapped-by-0 overlapped-by--1 overlapped-by--0
			same-time))
	 '1)
	((member pred '(between between-1 between-0 betwen--1 between--0
                        between-1-1 between-1-0 between-0-0 between-0-1))
	 '2)
	((member pred '(at-most-before at-least-before exactly-before
                        at-most-after at-least-after exactly-after))
	 '3)
	(t '4)
	)
  )

; classify-lits classifies the given literals, and returns a list
; of substitution orders and which literal to enter.
; Each item in the list is of the form (orderlist . number of lit to enter)
; Each item in orderlist is of the form 
;     (element number from 1st lit . element number from second)

(defun classify-lits (neg-p1 pred1 neg-p2 pred2 &aux
			     (pclass1 (classify-pred pred1))
			     (pclass2 (classify-pred pred2))
			     (which (determine-enter neg-p1 pclass1 
						     neg-p2))
			     ordlist)
  (unless (null which)
  (cond ((or (and (equal pclass1 '1)     ; before, after etc 
	          (equal pclass2 '1))    ; before, after etc 
	     (and (equal pclass1 '1)     ; before, after etc 
	          (equal pclass2 '3))    ; during relation
	     (and (equal pclass1 '3)     ; during relation
		  (equal pclass2 '1)))   ; before, after, etc 

	 ; each literal has only 2 args which can be unified
	 ; only two possible unifications - same direction and reversed

	 (setq ordlist (list (cons '((1 . 1)(2 . 2)) which)
			     (cons '((1 . 2)(2 . 1)) which))))

	((and (equal pclass1 '2)      ;between 
	      (equal pclass2 '2))

	 ; each lit has 3 args which may hbe unified with any of the other
	 ; lits three literals

	 (setq ordlist (list (cons '((1 . 1)(2 . 2)(3 . 3)) which)
			     (cons '((1 . 1)(2 . 3)(3 . 2)) which)
			     (cons '((1 . 2)(2 . 1)(3 . 3)) which)
			     (cons '((1 . 2)(2 . 3)(3 . 1)) which)
			     (cons '((1 . 3)(2 . 2)(3 . 1)) which)
			     (cons '((1 . 3)(2 . 1)(3 . 2)) which)))
	 )

	((and (member pclass1 '(1 3))  ; only 2 args to unify
	      (equal pclass2 '2))      ; 3 args to unify

	 ; there are 2 lits in the 1st lit which may unify with 3 in the 2nd

	 (setq ordlist (list (cons '((1 . 1)(2 . 2)) which)
			     (cons '((1 . 2)(2 . 1)) which)
			     (cons '((1 . 1)(2 . 3)) which)
			     (cons '((1 . 3)(2 . 1)) which)
			     (cons '((1 . 3)(2 . 2)) which)
			     (cons '((1 . 2)(2 . 3)) which)))
	 )

	((and (equal pclass1 '2)        ; 3 args to unify
	      (member pclass2 '(1 3)))  ; only 2 args to unify

	 ; there are 2 lits in the 2nd lit which may unify with 3 in the 1st

	 (setq ordlist (list (cons '((1 . 1)(2 . 2)) which)
			     (cons '((1 . 2)(2 . 1)) which)
			     (cons '((1 . 1)(3 . 2)) which)
			     (cons '((1 . 2)(3 . 1)) which)
			     (cons '((2 . 2)(3 . 1)) which)
			     (cons '((2 . 1)(3 . 2)) which)))
	 )

	)
  )
  ordlist
  )

; create-residue creates a literal of the given neg pred ard args

(defun create-residue (neg pred &rest args &aux
			    (reln (find-symbol (symbol-name pred) 'epilog)))
  (if (or (null pred)
	  (null args)
	  (eq (car args) (cadr args)))
      t
      (make-prop neg :pred reln :args args)
      )
  )

; diff-reln returns the difference between two relations , where reln1 subsumes
; reln2. 

(defun diff-reln (reln1 reln2)
  (if (member reln2 '(equal same-time))
      (more-strict reln1)
      'same-time)
  )

; incompatible-between splits the request into two requests, does them,
; and combines the results

(defun incompatible-between (negp reln a1 a2 a3 effort 
				 &optional (incompatible t) &aux
			    (testps (apply 'determine-split (split-time-pred reln)))
		            (test1 (apply 'build-pred (car testps)))
		            (test2 (apply 'build-pred (cadr testps)))
		            (ans1 (incompatible-test-pred negp test1 
					     (list a1 a2) effort incompatible))
		            (ans2 (incompatible-test-pred negp test2 
					     (list a1 a3) effort incompatible)))
  (if negp
      (cond ((or (equal ans1 t)
                 (equal ans2 t))
	     t)
	    ((equal ans1 nil)
	     ans2)
	    ((equal ans2 nil)
	     ans1)
	    (t (append ans1 ans2)))

      (cond ((or (null ans1)
                 (null ans2))
             nil)
	    ((equal ans1 t)
	     ans2)
	    ((equal ans2 t)
	     ans1)
	    (t (append ans1 ans2)))
		   
      )
  )

; incompatible-test-pred tests the other literal, now that one has been entered.

(defun incompatible-test-pred (negp reln args effort 
				   &optional (incompatible t) &aux
			      (pclass (classify-pred reln))
		              (a1 (car args))
		              (a2 (cadr args))
		              (a3 (caddr args)))
  (cond ((equal pclass '3) ;at-most-before, etc

	 ; for duration predicates, only evaluation done - cannot get
	 ; residues

	 (if (eq (eval-with-neg negp reln args effort) 
		 (if incompatible 'NO 'YES))
	     nil
	     t))

	((equal pclass '2)   ;between
	 ; separate into two requests
	 (incompatible-between negp reln a1 a2 a3 effort incompatible))


	; otherwise, compare this relation with the one between those
	; item sin the time graph. Depending on whether this literal
	; is negated or not, and this relationship, we can determine
	; a residue.

	(t (let* ((testreln (find-relation a1 a2 
					   *specialist-eval-effort*))
		  (creln (compare-time-preds reln testreln)))
		 (if negp
		     (case creln
		           (equivalent (not incompatible))
		           (subsumes (not incompatible))
		           (subsumed  ;should be difference <= & = -> <
			     (create-residue nil
		             (diff-reln reln testreln) a1 a2))
		           (intersects (if (member testreln '(equal same-time))
					   nil
					   (create-residue nil
					      (more-strict testreln) a1 a2)))
		           (disjoint incompatible)
		           (otherwise t)
		           )
		     (case creln
		           (equivalent incompatible)
		           (subsumes incompatible) 
		           (subsumed (create-residue nil reln a1 a2))
		           (intersects (create-residue nil 'same-time a1 a2))
		           (disjoint (not incompatible))
		           (otherwise t)
		           )
		     )
		 )
	   )
	)
  )

; enter-to-test just ensuere that no negations get passed to enter

(defun enter-to-test (negp pred arglist)
  (if negp
      nil
      (enter-time pred arglist))
  )

(defun residue-member (literal residue) 
  (or (equal (print-info literal) residue)
      (same-thing literal residue))
  )

; this just rules out the possiblity of getting a residue of
; (a before b) when one of the original literals was (a before b c)

(defun same-thing (literal residue &aux args1 args2)
  (unless (null residue)
	  (and (equal (getp literal 'pred) (prop-pred residue))
	       (equal (neg-p literal) (is-negated residue))
	       (equal (car (setq args1 (getp literal 'arguments)))
		      (car (setq args2 (prop-args residue))))
	       (equal (cadr args1) (cadr args2))
	      ))
  )


; incompatible-add checks to make sure that the residues are not the same as the
; original literals (without substitutions). If so, a subgoal item is
; added, otherwise, nothing is done.

(defun incompatible-add (lit-1 lit-2 res1 res2 eval1 eval2 subs1 subs2
			       &optional (incompatible t))
  (declare (ignore eval1 eval2))
  (unless (or (residue-member lit-1 res1)
	      (residue-member lit-1 res2)
	      (residue-member lit-2 res1)
	      (residue-member lit-2 res2))
          (when (traced-p 'epilog::time-test)
		    (print-line "    Time Specialist: Found " 
				(print-info lit-1 )
		     (if incompatible " incompatible" " compatible")
		     " with " (print-info lit-2))
		(print-line "        with residue: "
			    (cond (res1) (res2) (t "null")) )
		(when (or subs1 subs2)
		    (print-line "        with substitutions: " 
			       (print-subs (append subs1 subs2)))))
          (list
	  (make-comparison-info :lit1 lit-1 :lit2 lit-2
			    :subs1 subs1 :subs2 subs2
			    :residue1 res1
			    :residue2 res2
			    ; evaluations? 
			    ) )
	  )
  )

; incompatible-enter-and-compare checkpoints the net, enters one literal, and
; tests the other
; The net is then retracted to its previous state

(defun incompatible-enter-and-compare (lit-1 lit-2 entry eval1 eval2
				      neg-p1 reln1 newargs1
			              neg-p2 reln2 newargs2 subs1 subs2 effort 
				      &optional (incompatible t)
				      &aux
				      results cp resans)
  (unless (null entry)
   
  (setq cp (checkpoint "Time-CP"))
  (when (and (equal entry '2) (not (equal eval2 'no)))
 	(enter-to-test neg-p2 reln2 newargs2)
	(setq resans (incompatible-test-pred neg-p1 reln1 newargs1 effort
					     incompatible))
	(cond ((equal resans t)
	       nil)
	      (t (setq results (append results
		       (incompatible-add lit-1 lit-2 resans nil nil nil
				    subs1 subs2 incompatible))))
	      )
	)
  (when (and (equal entry '1) (not (equal eval1 'no)))
  	(enter-to-test neg-p1 reln1 newargs1)
	(setq resans (incompatible-test-pred neg-p2 reln2 newargs2 effort
					     incompatible))
	(cond ((equal resans t)
	       nil)
	      (t (setq results (append results
		       (incompatible-add lit-1 lit-2 nil resans nil nil
				    subs1 subs2 incompatible))))
	      )
	)
  (retract cp :remove t)
  )
  results
  )

; test-residue does the necessary substitutions, and then determines
;compatibility or incompatibility of the resulting literals

(defun test-residue (incompatible lit-1 lit-2 ord-entry
			     neg-p1 reln1 arglist1
			     neg-p2 reln2 arglist2 effort &aux
 	                     (ordlist (car ord-entry))
 	                     (entry (cdr ord-entry))
			     (newargs1 arglist1)
			     (newargs2 arglist2)
			     (eval1 'unknown)
			     (eval2 'unknown)
			     newlit1 newlit2 res
	                     ulist subs1 subs2 s1)
  (unless (null ordlist)
          (setq ulist (unify-lists arglist1 arglist2 ordlist 'unifiable))

          (unless (null ulist)
                  (setq subs1 (car ulist))
	          (setq subs2 (cdr ulist))
	          (setq newlit1 (normalize-wff (substitute-prop lit-1 subs1)))
	          (setq newlit2 (normalize-wff (substitute-prop lit-2 subs2)))
		  ))

  (unless (or (null newlit1)
	      (null newlit2))
	  (when (traced-p 'epilog::time-test)
		(print-line "        Time Specialist: Comparing " (getp newlit1 'print)
			    " with " (getp newlit2 'print)))
          
	  (setq newargs1 (spec-eval-args newlit1))
	  (setq newargs2 (spec-eval-args newlit2))
	  (when (or (equal (car newargs1) (cadr newargs1))
	            (equal (car newargs2) (cadr newargs2)))
		(return-from test-residue nil))

	  (if (getp newlit1 'epilog::prob) 
		       (progn (setq eval1 'YES)
			      (when (traced-p 'epilog::time-eval)
				  (print-line "    Time Specialist: Evaluated " 
					  (getp newlit1 'print) " to " eval1)))
	          (setq eval1 (eval-with-neg neg-p1 reln1 newargs1 effort)))
	  (if (getp newlit2 'epilog::prob) 
		       (progn (setq eval2 'YES)
			      (when (traced-p 'epilog::time-eval)
				  (print-line "    Time Specialist: Evaluated " 
					  (getp newlit2 'print) " to " eval2)))
	          (setq eval2 (eval-with-neg neg-p2 reln2 newargs2 effort)))

          (when (equal eval1 'unknown)
                (setq subs1 (union subs1 
		                (setq s1 (arg-sub-list newargs1 
						 (getp newlit1 'arguments) 1))))
	        (setq newlit1 (normalize-wff (substitute-prop newlit1 s1)))
	        )
          (when (equal eval2 'unknown)
                (setq subs2 (union subs2 
		              (setq s1 (arg-sub-list newargs2 
						 (getp newlit2 'arguments) 2))))
	        (setq newlit2 (normalize-wff (substitute-prop newlit2 s1)))
	        )

	  ; first evaluate both literals

	  (cond ((setq res 
                     (incompatible-with-eval lit-1 lit-2 eval1 eval2 subs1 subs2
		                        (traced-p 'epilog::time-test)
		                        "    Time Specialist:" incompatible))
		 res)

		; if this gives no answer, enter one and test the other
		; ensure that we don't enter contradictory info into time graph

		((and (member eval1 '(YES NO)) (member eval2 '(YES NO))) nil)

		((equal (if (equal entry 1) eval1 eval2) 'no) nil)

		((or (equal eval1 'NO) (Equal eval2 'NO)) nil)

		(t (incompatible-enter-and-compare 
				     lit-1 lit-2 entry eval1 eval2
			              neg-p1 reln1 newargs1
			              neg-p2 reln2 newargs2
			              subs1 subs2 effort incompatible)
		       )
		)
	  )
  )

; if arg is a constant, it must be of correct sort - if variable, must be
; of correct sort, or have no sort

(defun okarg (arg &aux
		  (s (find-sort arg)))
  (if (check-constant arg)
      (if (member s '(epilog::episode epilog::time))
	  t
	  nil)
      (or (null s)
	  (if (member s '(epilog::episode epilog::time))
	      t
	      nil))
      )
  )

(defun check-arg-sorts (pred arglist)
  (cond ((member pred '(at-most-before at-most-after at-least-before 
			at-least-after exactly-before exactly-after))
	 (if (and (okarg (car arglist))
		  (okarg (cadr arglist))
		  (number-arg (caddr arglist)))
	     t
	     nil))
	(t  (eval (append (list 'and) (mapcar #'(lambda (a) 
					  (okarg a))
				arglist))))
	)
  )

; incompatible-lits tries to determine whether lit1 and lit2 are incompatible
; it tries all possible substitutions to find out

(defun incompatible-lits (netname lit1 lit2 neg-p1 pred1 arglist1 
		   	   neg-p2 pred2 arglist2 effort 
			   &optional (incompatible t) &aux
		           (reln1 (find-symbol (symbol-name pred1) 'time))
		           (reln2 (find-symbol (symbol-name pred2) 'time))
	                   (all-constants (not (or 
						(non-constant-arg arglist1)
				                (non-constant-arg arglist2)
						)))
			   )

  (assert (eq *current-netname* netname))
  ;; I'm worried by the fact that there used to be two ways for the
  ;; netname information to get into the specialists, via the netname
  ;; argument  and via *current-netame*.  I want to be sure that they're
  ;; consistent. -ANK
  
    (when (and (check-arg-sorts reln1 arglist1)
	     (check-arg-sorts reln2 arglist2))

  ; see if arguments unifiable, and in what order

  ; enter one literal, then test the relationship for the
  ; arguments in the other one against the net relationship

  (when (traced-p 'epilog::time-test)
	(print-line "    Time Specialist: Trying to determine "
                (if incompatible "incompatiblity" "compatibility") " of  " 
		(print-info lit1) " and " (print-info lit2))
	(if (not (equal *current-netname* '/)) 
	    (print-it "    in subnet " *current-netname*)) 
	)

  (if all-constants
      (let ((eval1 (eval-with-neg neg-p1 reln1 arglist1 effort))
	    (eval2 (eval-with-neg neg-p2 reln2 arglist2 effort)))
	   (cond ((incompatible-with-eval lit1 lit2 eval1 eval2 nil nil 
		     (traced-p 'epilog::time-test)
		     "    Time Specialist:" incompatible))
		 (t
                   (incompatible-enter-and-compare
		                              lit1 lit2 
		                              (determine-enter 
					         neg-p1 (classify-pred reln1) 
						 neg-p2)
					      eval1 eval2
			                      neg-p1 reln1 arglist1
			                      neg-p2 reln2 arglist2
			                      nil nil effort
					      incompatible)
		   )))

      (let* ((ord-entry (classify-lits neg-p1 reln1 neg-p2 reln2)))
	    (if (null ord-entry)
	        nil
                (let (r)
		     (mapcar #'(lambda (o) 
				       (setq r (append r
					       (test-residue incompatible
					 	  lit1 lit2 o 
						  neg-p1 reln1 arglist1
					          neg-p2 reln2 arglist2 
						  effort))) )
				      ord-entry)
		     r)
		)
	    )
      ))
  )

; compatible-lits tries to determine whether lit1 and lit2 are compatible.
; it tries all possible substitutions to do this.

(defun compatible-lits (netname lit1 lit2 neg-p1 pred1 arglist1 
		   	   neg-p2 pred2 arglist2 effort )
  (assert (eq *current-netname* netname))
  ;; I'm worried by the fact that there used to be two ways for the
  ;; netname information to get into the specialists, via the netname
  ;; argument  and via *current-netame*.  I want to be sure that they're
  ;; consistent. -ANK
  
  (incompatible-lits netname lit1 lit2 neg-p1 pred1 arglist1 neg-p2 pred2 
		     arglist2 effort nil)
  )

