;       Last change    August 1993
(in-package time)
;                    ENTRY Routines

; inconsistent returns t if the two points would be inconsistent if the
; relation reln were added - this is used to ensure that if evaluation came
; back with unknown (therefore ok to enter), the net remains consistent after
; the entry
;  for example, if a <= b and we add a >= b, evaluation is unknown, but
;  to remain consistent, a must = b

(defun inconsistent (f l reln &aux
		       (oldreln (find-point-reln f l 
					 *specialist-entry-effort*))) 
  (case reln
	(before 
	 (if (member oldreln '(equal after after-0 before-0))
	     t
	     nil)
	 )
	(after
	 (if (member oldreln '(equal before before-0 after-0))
	     t
	     nil)
	 )
	(otherwise nil)
	)
  )

;; Update-point resets the time point array for pt to point to newnode

(defun update-point (ptname newnode)
  (assert (writeable-p newnode))
  (change-field 'time-node-alternate-names
		newnode
		(pushnew ptname (time-node-alternate-names newnode)))
  ;; It seems to me that it would be a bug to use pushnew in a
  ;; change-field value.  The pushnew gets evaluated in preparation
  ;; for calling change-field, so change-field will store the new value
  ;; on the rewind stack as the old value.  I haven't yet tested this to
  ;; see if I'm right. -ANK
  (change-hash ptname
	       (time-net-time-graph (writeable-net))
	       newnode))

; Update-first checks to see if the new point pt is earlier on its
; chain than the previous first point. If so, it updates the first
; indicator.

(defun update-first (pt &aux
			    (meta (time-node-chain pt))
			    (first (meta-node-first meta)))
  (assert (writeable-p pt))
  (assert (writeable-p meta))      
  (if (or (null first)
          (< (time-node-psuedo pt) (time-node-psuedo first)))
      (change-field 'meta-node-first meta pt)
      nil)
  )

; First desc returns the first descendant of point on the same chain

(defun first-desc (point &aux
                         (dlist (time-node-descendants point)))
  (if (null dlist)
      point
      (time-point (time-link-to-point-name (car dlist)))
      )
   )

; First anc returns the nearest ancestor of point on the same chain

(defun first-anc (point &aux
                         (alist (time-node-ancestors point))
			 (item (last alist)))
  (if (null item)
      point
      (time-point (time-link-from-point-name (car item)))
      )
   )

; Check-first is used when a point is being replaced (through equal)
; that is the first point on its chain.
; In that case, the first indicator is set to point's first descendant,
; and the minimum psuedo propagated if necessary.
; Note that if this were happening within the same chain, the first point would
; be kept, and the others made equal to it, so this situation would not arise.

(defun check-first (point &aux
                        (mn (time-node-chain point)))
  (assert (writeable-p point))
  (when (equal (time-node-name point) 
               (time-node-name (meta-node-first mn)))
	(let* ((newp (first-desc point)))
              (change-field 'meta-node-first mn newp)
	      (when (equal (min-psuedo newp) (min-psuedo point))
		    (change-field 'time-node-min-point-name newp
                        (time-node-name newp))
		    (propagate-min (first-desc newp) newp)
		    )
	      )
        )
   )

; Add-single adds a single point to the net on a new chain

(defun add-single (ptname)
  (let ((tpt (make-time-node :name ptname
			     :chain (newchain)
			     :psuedo newpsuedo)))
    (change-hash ptname
		 (time-net-time-graph (writeable-net))
		 tpt)
    (update-first tpt)
    tpt))

; Collapse-nodes collapses two nodes into one, to make them equal.
; It first copies all the link information from the second point to the 
; first, and then makes the second point actually point to the first.

(defun collapse-nodes (t1 t2) 
  (assert (writeable-p t1))
  (assert (writeable-p t2))
  (if (traced-p 'epilog::time-point-entry)
      (print-line "    Time-Specialist: Collapsing points " 
		       (time-node-name t1) " and " 
		       (time-node-name t2)))

  (copy-links t2 t1)
; maybe notify spec interface of changes here?
  (dolist (p (time-node-alternate-names t2))
	  (update-point p t1))
  (change-field 'time-node-alternate-names t1 (union 
					(time-node-alternate-names t2)
					(time-node-alternate-names t1)))
  (update-point (time-node-name t2) t1)
  )

; Get-path returns a path of all the points in the same chain, in order
; from t1 to t2, inclusive.

(defun get-path (t1 t2 &aux
		       (next (time-point
			      (time-link-to-point-name
			       (car (time-node-descendants t1))))))
  (cond ((eq t1 t2)
	 (list t1))
	((null next)
	 (print-error "(get-path " t1 t2 "): no such path exists."))
	(t (append (list t1)
		   (get-path next t2)))))

; Collapse-chain is used to make two points on the same chain equal -
; all points between must be equal as well.

(defun collapse-chain (t1 t2 &aux
			   (min1 (min-psuedo t1))
			   (minptname (time-node-min-point-name t2))
			   (min2 (min-psuedo t2))
			   (max1 (max-psuedo t1))
			   (maxptname (time-node-max-point-name t2))
			   (max2 (max-psuedo t2))
			   (absmin1 (time-node-absolute-min t1))
			   (absmin2 (time-node-absolute-min t2)))

  (assert (writeable-p t1))
  (assert (writeable-p t2))

  ; collapse all points between

  (dolist (item (cdr (get-path t1 t2)))
	  (collapse-nodes t1 item)
	  )

  ; max and min must be most constrained of the points

  (when (and (not (null min2))
	     (or (null min1)
	         (> min2 min1)))
	(change-field 'time-node-min-point-name t1 minptname)
	(propagate-min (first-desc t1) (time-point minptname))
	)
  (when (and (not (equal max2 t))
	     (or (null max1)
	         (< max2 max1)))
	(change-field 'time-node-max-point-name t1 maxptname)
	(propagate-max (first-anc t1) (time-point maxptname))
	)

  ; update absolute time minimum if necessary - no need to propagate
  ; because propagation of min goes forward, and this would already 
  ; have been done. Max would have been propagated back originally if
  ; max of pt2 were < max of pt1
    
  (when (or (null absmin1)
	    (equal (test-answer 'after (compare-abs absmin2 absmin1)) 'yes))
	(change-field 'time-node-absolute-min t1 absmin2)
	)
  )

; Collapse-xchain makes two nodes on separate chains equal - it first
; checks to see if t2 is first on its chain, and if so, replaces the first 
; by t2's first descendant.
; Any links t2 has are copied to t1, and t2 is replaced by t1 in the
; time graph

(defun collapse-xchain (t1 t2 &aux
			   (absmin2 (time-node-absolute-min t2))
			   (absmax2 (time-node-absolute-max t2))
			   )
  (assert (writeable-p t1))
  (assert (writeable-p t2))
  (check-first t2)
  (copy-links t2 t1)

  (if (traced-p 'epilog::time-point-entry)
      (print-line  "    Time-Specialist: Collapsing points " 
		       (time-node-name t1) " and " 
		       (time-node-name t2)))

  ; copy absolute time
  (update-absolute-min t1 absmin2)
  (update-absolute-max t1 absmax2)

  (dolist (p (time-node-alternate-names t2))
	  (update-point p t1))
  (change-field 'time-node-alternate-names t1 (union
					(time-node-alternate-names t2)
					(time-node-alternate-names t1)))
  (update-point (time-node-name t2) t1)
  )

; Add-equal makes two points equal.  pt1 is either a node structure,
; or a point name if no structure yet exists for that naem; pt2 is the
; node structure for the second point, which must already exist.  If
; name1 is new, it is just made another name for pt2.  If both exist,
; information from pt2 is copied to pt1, and then pt2 is replaced by
; pt1.

(defun add-equal (pt1 pt2)

    ;; if pt1 is new, just make it point to pt2

    (if (not (time-node-p pt1))
	(update-point pt1 (writeable-point pt2))

      ; otherwise things are a bit more complicated
      ; We need to copy relevant info from pt2 to pt1 and replace pt2 by pt1

      (progn

	     ; check to make sure that these points could be equal - 
	     ; otherwise could get loops
	     ; this test only for those in same chain - should actually
	     ; do consistency search for others

	     (cond ((and (on-same-chain pt1 pt2)
			 (not (possibly-equal pt1 pt2))) nil)

		   ; see if they are already equal - no need to do anything

	           ((equal pt1 pt2)
		    nil)

		   ; if pt2 has not been replaced by any other point, copy
		   ; its information to pt1 and replace pt2

                   (t
		    (let ((pt1 (writeable-point pt1))
			  (pt2 (writeable-point pt2)))
		      (if (on-same-chain pt1 pt2)
			  (collapse-chain pt1 pt2)
			(collapse-xchain pt1 pt2))))))))

; Add-strictness modifies the max psuedo of the first point and the min psuedo 
; of the last point to ensure that they cannot be equal. This information is 
; propagated if necessary.

(defun add-strictness (f l &aux
			 (oldmin (min-psuedo l))
			 (oldmax (max-psuedo f)))

  ; set minimum of l to be f, if this is more restrictive

  (when (or (null oldmin)
	    (> (time-node-psuedo f) oldmin))
    (setq l (writeable-point l))
    (change-field 'time-node-min-point-name l (time-node-name f))
    (propagate-min l f)
    )

  ; set maximum of f to be l, if this is more restrictive

  (when (or (equal oldmax t)
	    (< (time-node-psuedo l) oldmax))
    (setq f (writeable-point f))
    (change-field 'time-node-max-point-name f (time-node-name l))
    (propagate-max f l)
    )
  )

; Propagate-min propagates minimum psuedo time forward along descendants 
; of a chain until it reaches a minimum greater than it

(defun propagate-min (frompt newmin &aux
			     (dlist (time-node-descendants frompt))
			     (item (car dlist)))
  (unless (null newmin)
    (let ((psuedo (min-psuedo frompt)))
      (when (or (null psuedo)
		(>= (time-node-psuedo newmin) psuedo))
	(setq frompt (writeable-point frompt))
	(change-field 'time-node-min-point-name frompt
		      (time-node-name newmin))
	(unless (null item) 
	  (propagate-min (time-point
			  (time-link-to-point-name item))
			 newmin)
	  )
	)
      )
    )
)
  
; Propagate-max propagates maximum psuedo time backward along ancestors 
; of a chain until it reaches a maximum less than it

(defun propagate-max (frompt newmax &aux
			     (alist (time-node-ancestors frompt))
			     (item (car (last alist))))
  (unless (equal newmax t)
    (let ((psuedo (max-psuedo frompt)))
      (when (or (equal psuedo t)
	            (<= (time-node-psuedo newmax) psuedo))
	(setq frompt (writeable-point frompt))
	(change-field 'time-node-max-point-name frompt
		      (time-node-name newmax))
	(unless (null item) 
	  (propagate-max (time-point
			  (time-link-from-point-name item))
			 newmax)
	  )
	)
      )
    )
)

; Add-before-same-chain sets the chain of point f to be the same as l's,
; and sets its psuedo time to be before l's.

(defun add-before-same-chain (f l strictfl)
  (assert (writeable-p f))
  (assert (writeable-p l))
  (setf (time-node-chain f) (time-node-chain l))
  (setf (time-node-psuedo f) (psuedo-before l))
  (if (strict strictfl)

      ; if the arc between is strict, f's maximum is l,
      ; l's minimum is f, and this is propagated

      (let ()
 	   (setf (time-node-max-point-name f) (time-node-name l))
	   (change-field 'time-node-min-point-name l (time-node-name f))
	   (propagate-min l f)
	   )

      ; otherwise, f's maximum is l's maximum

      (setf (time-node-max-point-name f) (time-node-max-point-name l))
      )
  (update-first f)
  )

; Add-after-same-chain sets the chain of point l to be the same as f's,
; and sets its psuedo time to be after f's.

(defun add-after-same-chain (l f strictfl)
  (assert (writeable-p l))
  (assert (writeable-p f))
  (setf (time-node-chain l) (time-node-chain f))
  (setf (time-node-psuedo l) (psuedo-after f))
  (if (strict strictfl)

      ; if the arc between is strict, l's minimum is f,
      ; f's maximum is l, and this is propagated

      (let ()
 	   (setf (time-node-min-point-name l) (time-node-name f))
	   (change-field 'time-node-max-point-name f (time-node-name l))
	   (propagate-max f l)
	   )

      ; otherwise, l's minimum is f's minimum

      (setf (time-node-min-point-name l) (time-node-min-point-name f))
      )
  )

; Add-between-same-chain sets the chain of m to be the same as f and l,
; sets its psuedo time to be between

(defun add-between-same-chain (m f l strictfm strictml)
  (assert (writeable-p m))
  (assert (writeable-p f))
  (assert (writeable-p l))
  (setf (time-node-chain m) (time-node-chain f))
  (setf (time-node-psuedo m) (psuedo-between f l))

  ; set m's minimum

  (if (strict strictfm)

      ; if the arc between f and m  strict, m's minimum is f,
      ; f's maximum is m, and this is propagated

      (let ()
       (setf (time-node-min-point-name m) (time-node-name f))
       (change-field 'time-node-max-point-name f (time-node-name m))
       (propagate-max f m)
       (propagate-min l f)
       )

      ; otherwise, m's minimum is f's minimum

      (setf (time-node-min-point-name m) (time-node-min-point-name f))
      )

  ; set m's maximum

  (if (strict strictml)

      ; if the arc between m and l strict, m's maximum is l,
      ; l's minimum is m, and this is propagated

      (let ()
       (setf (time-node-max-point-name m) (time-node-name l))
       (change-field 'time-node-min-point-name l (time-node-name m))
       (propagate-min l m)
       (propagate-max f l)
       )

      ; otherwise, m's maximum is l's maximum

      (setf (time-node-max-point-name m) (time-node-max-point-name l))
      )
  )

; Add-on-new-chain sets the chain of point to be a new chain, and sets
; the first indicator for that new chain

(defun add-on-new-chain (point)
  (assert (writeable-p point))
  (setf (time-node-chain point) (newchain))
  (setf (time-node-psuedo point) newpsuedo)
  (update-first point)
  )


; Add-before adds first before last. If first doesn't exist, it is added on
; the same chain as last if possible. Otherwise it starts a new chain.
; A link is added between them reflecting the strictness.

(defun add-before (firstname lastname strict &aux
					     (f (time-point firstname))
					     (l (writeable-point lastname)))
  ; if first is new, create it

  (if (null f)
      (progn
        (setq f (make-time-node :name firstname))
        (update-point firstname f)
        (if (first-in-chain l)
	    (add-before-same-chain f l strict)
            (add-on-new-chain f)
	    ))
    (setq f (writeable-point f)))

  ; if they already exist and are on same chain, adjust psuedo-times
  ; if necessary to reflect strictness.

  (if (and f strict (on-same-chain f l))
      (add-strictness f l))

  ; add the link

   (add-link f l strict)

  ; update absolute times

   (update-absolute-max f (time-node-absolute-max l))
   (update-absolute-min l (time-node-absolute-min f))
   )

; Add-after adds last after first. If last doesn't exist, it is added on
; the same chain as first if possible. Otherwise it starts a new chain.
; A link is added between them reflecting the strictness.

(defun add-after (lastname firstname strict &aux
					    (f (writeable-point firstname))
					    (l (time-point lastname))) 

  ; if last is new, create it

  (if (null l)
      (progn
        (setq l (make-time-node :name lastname))
        (update-point lastname l)
        (if (last-in-chain f)
	    (add-after-same-chain l f strict)
            (add-on-new-chain l)
	    ))
    (setq l (writeable-point l)))

  ; if they already exist and are on same chain, adjust psuedo-times
  ; if necessary to reflect strictness.

  (if (and l strict (on-same-chain f l))
      (add-strictness f l))

  ; add the link

  ;; This doesn't check for an existing link before adding a new one.
  ;; That seems wrong to me, but I haven't changed it. -ANK

   (add-link f l strict)

  ; update absolute times

   (update-absolute-min l (time-node-absolute-min f))
   (update-absolute-max f (time-node-absolute-max l))
   )

; Adjacent returns t if first and last are next to each other on the chain
; with no intervening points. Otherwise it returns nil.

(defun adjacent (first last)
  
 ; if first descendant of first is last, they are adjacent
 ; otherwise there is some intervening point so they are not
  
  (equal (time-node-name last) (time-node-name (first-desc first)))
) 

; Add-between adds middle between first and last. If middle is new
; it tries to add it on the same chain as first and/or last to minimize
; chains. If this is not possible, a new chain is started.
; Links are created between fist and middle and middle and last reflecting
; the strictness given.

(defun add-between (middlename firstname lastname strictfm strictml &aux
		       (f (writeable-point firstname))
		       (m (time-point middlename))
		       (l (writeable-point lastname)))
  ;; if middle is new, create it, unless it's just being set equal to
  ;; one of the other points.
  (if (null m)
      (cond ((meets strictfm)
	     (add-equal middlename f)
	     (setq m f))
	    ((meets strictml)
	     (add-equal middlename l)
	     (setq m l))
	    (t (setq m (make-time-node :name middlename))
	       (update-point middlename m)
	       (cond ((on-same-chain f l)
		      
		      ;; if adjacent, add on same chain, else new chain
		      
		      (if (adjacent f l)
			  (add-between-same-chain m f l strictfm strictml)
			(add-on-new-chain m)))
		     
		     ;; if first is at end of chain, add middle after it
		     
		     ((last-in-chain f)
		      (add-after-same-chain m f strictfm))
		     
		     ;; if last is at beginning of its chain,
		     ;; add middle before it
		     
		     ((first-in-chain l)
		      (add-before-same-chain m l strictml))
		     
		     ;; otherwise add it on a new chain
		     
		     (t (add-on-new-chain m))
		     )
	       )
	    )
    (setq m (writeable-point m))
    )
   
  ;; if strictness indicates that first = middle, and middle wasn't
  ;; just set to first above, make them equal; otherwise add the link
  ;; between first and middle

  (if (meets strictfm)
      (if (not (eq f m))
	  (add-equal m f))
    (add-link f m strictfm))
   
  ;; if strictness indicates that middle = last, and middle wasn't
  ;; just set to first above, make them equal; otherwise add the link
  ;; between middle and last

  (if (meets strictml)
      (if (not (eq m l))
	  (add-equal m l))
    (add-link m l strictml))

  ; update absolute times

   (update-absolute-min m (time-node-absolute-min f))
   (update-absolute-max m (time-node-absolute-max l))
   (update-absolute-max f (time-node-absolute-max m))
   (update-absolute-min l (time-node-absolute-min m))
   )

; Check-equal ensures that the call to add-equal will have the correct
; arguments 

(defun check-equal (name1 name2)
  (let ((pt1 (time-point name1))
	(pt2 (time-point name2)))
    (cond ((null pt1)

	   ;; if pt1 is new and pt2 is new, add pt2 as a new point

	   (when (null pt2)
	     (setq pt2 (add-single name2)))

	   ;; Now arguments are ok for add-equal

	   (if pt1
	       (add-equal pt1 pt2)
	     (add-equal name1 pt2)))
	  ;; if pt1 does exist and pt2 doesn't, reverse the order of
	  ;; the arguments (equal is symmetric)

	  ((null pt2)
	   (add-equal name2 pt1))

	  ;; otherwise, both exist 

	(t
	 ;; if both are on the same chain, put the earlier one first in
	 ;; the call to prevent having to change the first indicator for
	 ;; that chain

	 (if (and (on-same-chain pt1 pt2)
		  (> (time-node-psuedo pt1)
		     (time-node-psuedo pt2)))
	     (add-equal pt2 pt1)
	   (add-equal pt1 pt2)
	   )
	 )
	)
    )
  )

; Check-before ensures that the arguments are correct for add-before

(defun check-before (firstname lastname strictfl &aux
					 (sfl strictfl))
  ; first check to see if this is potentially inconsistent - if so, 
  ; set strictness so they will be made equal instead

  (if (and (notstrict strictfl)
	   (inconsistent firstname lastname 'before))
      (setq sfl '0)
      )

  ; if the strictness indicates meets (equal) make the points equal

  (cond ((meets sfl) (check-equal firstname lastname))

	; if last doesn't exist use add-after, since it allows the end
	; point to be new

        ((null (time-point lastname))

	 ; if first doesn't exist either, add it

         (when (null (time-point firstname))
	   (add-single firstname))

         (add-after lastname firstname sfl))

	; otherwise ready for add-before

        (t (add-before firstname lastname sfl))
	)
  )
  

; Check-after ensures that the arguments are correct for add-after

(defun check-after (lastname firstname strictfl &aux
			(sfl strictfl))
  ; first check to see if this is potentially inconsistent - if so, 
  ; set strictness so they will be made equal instead

  (if (and (notstrict strictfl)
	   (inconsistent lastname firstname 'after))
      (setq sfl '0)
      )

  ; if the strictness indicates meets (equal) make the points equal

  (cond ((meets sfl) (check-equal lastname firstname))

	; if first doesn't exist use add-before, since it allows the start
	; point to be new

        ((null (time-point firstname))

	 ; if last doesn't exist either, add it

         (when (null (time-point lastname))
               (add-single lastname))

         (add-before firstname lastname sfl))

	; otherwise ready for add-after

        (t (add-after lastname firstname sfl)))
  )

; Handle-between tries to minimize the number of chains required by
; using the appropriate combination of add-before, after and between

(defun handle-between (middlename firstname lastname strictfm strictml &aux
			   (sfm strictfm)
			   (sml strictml))
 
  ; if strictness indicates first and middle equal, make them equal
  ; and then try to add middle before last

  (when (and (notstrict (combine-strict sfm sml))
	     (inconsistent lastname firstname 'after))
        (setq sfm '0)
        (setq sml '0))

  (cond ((meets sfm)    ; first = middle
	 (check-equal middlename firstname)
	 (check-before middlename lastname sml))

	; if strictness indicates middle = last, make them equal
	; then try to add middle after first

	((meets sml)    ; last = middle
 	 (check-equal middlename lastname)
	 (check-after middlename firstname sfm))

	; otherwise use add-between

	(t (add-between middlename firstname lastname sfm sml))
	)
  )

 ; 
 ; Check-Between - prepare to call add-between which requires that
 ; first and last exist. If they don't, add them minimizing the number
 ; of chains to be created

(defun check-between (middlename firstname lastname strictfm strictml &aux
			  (sfm strictfm)
			  (sml strictml))
  ; first check to see if this is potentially inconsistent - if so, 
  ; set strictness so they will be made equal instead

  (if (and (notstrict strictfm)
	   (inconsistent middlename firstname 'after))
      (setq sfm '0))
  (if (and (notstrict strictml)
	   (inconsistent middlename lastname 'before))
      (setq sml '0))
  ;
  ; If middle doesn't exist
  ;
  (cond ((null (time-point middlename))

         ; make sure first and last exist, and if not, add them
         ; then add middle between

         (check-before firstname lastname (combine-strict sfm sml))
	 (handle-between middlename firstname lastname sfm sml)
         )

        ; middle already exists
        ; if middle is at either extreme of its chain, use before and
	; after to ensure that if first or last don't exist, the minimum
	; number of chains are created

	((or (first-in-chain (time-point middlename))
	     (last-in-chain (time-point middlename)))
	 (check-before firstname middlename sfm)
	 (check-after lastname middlename sml))

	; otherwise, if either first or last is missing, adding last
	; before first using check-after (minimize chains again)
	; and then add middle between

	(t (check-after lastname firstname (combine-strict sfm sml))
	   (handle-between middlename firstname lastname sfm sml)
	   )
	)
  )
 
; Enter-point enters a particular relation between time points. All
; arguments are points at this stage (no events)

(defun enter-point (pt1 stem pt2 pt3 s1 s2)
  (when (or (equal stem 'between)
	    (not (equal pt1 pt2)))
  (if (traced-p 'epilog::time-point-entry)
      (print-line "    Time Specialist: Entering point relation " pt1 
		     " " (build-pred stem s1 s2) " " pt2 " " 
		 (if pt3 (print-info pt3) " "))
      )
  (if (null stem)
      (add-single pt1)
      (case stem
            ((equal same-time)
	     (check-equal pt1 pt2))
            (before 
	     (check-before pt1 pt2 s1))
            (after 
	     (check-after pt1 pt2 s1))
            (between 
	     (check-between pt1 pt2 pt3 s1 s2))
	    )
      )
  )
  )

; add-event enters a link between the start and end point sof an event if
; one doesn't already exist

(defun add-event (e)
  (enter-point (event-node-start e) 'before (event-node-end e) nil nil nil)
  )

; enter-duration-reln enters the relation suggested by reln and also the
; duration bound (min or max)

(defun enter-duration-reln (a1 reln a2 a3)
  (case reln 
	(at-least-before
	 (enter-reln a1 'before a2 nil '1 nil)
	 (enter-duration-min a1 a2 a3)
	 )
	(at-most-before
	 (enter-reln a1 'before a2 nil '1 nil)
	 (enter-duration-max a1 a2 a3)
	 )
	(exactly-before
	 (enter-reln a1 'before a2 nil '1 nil)
	 (enter-duration-min a1 a2 a3)
	 (enter-duration-max a1 a2 a3)
	 )
	(at-least-after
	 (enter-reln a1 'after a2 nil '1 nil)
	 (enter-duration-min a2 a1 a3)
	 )
	(at-most-after
	 (enter-reln a1 'after a2 nil '1 nil)
	 (enter-duration-max a2 a1 a3)
	 )
	(exactly-after
	 (enter-reln a1 'after a2 nil '1 nil)
	 (enter-duration-min a2 a1 a3)
	 (enter-duration-max a2 a1 a3)
	 )
	)
  )

; enter handles the entry of a proposition  (pred applied to arglist)
; into the appropriate time graph (indicated by netname). 
  
(defun enter (netname lit negp pred arglist &aux
		      (reln (find-symbol (symbol-name pred) 'time))
		      (a2 (cadr arglist)))

  (declare (ignore lit))
  (assert (eq *current-netname* netname))
  ;; I'm worried by the fact that there used to be two ways for the
  ;; netname information to get into the specialists, via the netname
  ;; argument and via *current-netame*.  I want to be sure that they're
  ;; consistent. -ANK
  
  (if (or negp
	  (null arglist)
	  (non-constant-arg arglist)
	  (if (equal reln 'has-duration)
	      (not (member (find-sort a2) '(epilog::number epilog::integer)))
	      (not (member (find-sort a2) '(epilog::time epilog::episode)))))
      nil
      (enter-time reln arglist)
      )
  )

(defun enter-time (reln arglist &aux
		      (plist (split-time-pred reln))
		      (stem (stem-of plist))
		      (s1 (first-strict plist))
		      (s2 (second-strict plist))
		      (a1 (car arglist))
		      (a2 (cadr arglist))
		      (a3 (caddr arglist))
		      (enter-res nil))


  (unless (null arglist)

  (when (traced-p 'epilog::time-entry)
        (if (null a3)
            (print-line "    Time Specialist: Entering " 
			   (print-info a1) " " reln " " (print-info a2))
            (print-line "    Time Specialist: Entering " 
			   (print-info a1) " " reln " " (print-info a2) " " 
			   (print-info a3)) 
            )
	(if (not (equal *current-netname* '/))
	    (print-it "    into subnet " *current-netname*))
      )

  (when (event a1)
    (prepare-event a1)) ; why isn't add-event called here? -ANK
  (when (event a2)
    (prepare-event a2)) ; ditto
  (case stem
	((before after during overlaps overlapped-by contains)
	 (when (or (null a3)
		   (equal (find-sort a3) 'epilog::episode))
               (unless (null a3)
		 (add-event (prepare-event a3)))
               (if (or (absolute-time a1)
	               (absolute-time a2))
		   (if (absolute-time a2)
                       (enter-absolute a1 stem a2 a3 s1 s2)
		       (enter-absolute a2 (inverse-reln stem) a1 a3 s1 s2))
	           (enter-reln a1 stem a2 a3 s1 s2))
	       (setq enter-res t)
	       ))
	((equal same-time)
	 (if (absolute-time a1)
	     (enter-absolute a2 stem a1 nil nil nil)
	     (if (absolute-time a2)
	         (enter-absolute a1 stem a2 nil nil nil)
	         (enter-reln a1 stem a2 nil nil nil))
	     )
	 (setq enter-res t))
	(between
	 (when (member (find-sort a3) '(epilog::time epilog::episode))
	       (when (event a3)
	             (prepare-event a3))
	       (if (or (absolute-time a1)
	               (absolute-time a2)
		       (absolute-time a3))
	           (enter-abs-between a1 stem a2 a3 s1 s2)
	           (enter-between a1 stem a2 a3 s1 s2))
	       (setq enter-res t)
	       )) 
	((at exactly)
	 (unless (or (absolute-time a1)
		     (absolute-time a2)
		     (not (number-arg a3)))
	         (enter-duration-reln a1 reln a2 a3)
		 (setq enter-res t))
	 )
	(has-duration
	 (unless (or (absolute-time a1) (not (event a1))
		     (not (number-arg a2)))
		 ; enter start exactly before end with duration as elapsed
		 (enter-duration-reln (get-start a1) 'exactly-before
				      (get-end a1) a2)
		 (setq enter-res t)
		 )
	 )
	(otherwise
	 (print-line "Temporal relation " reln " not implemented"))
	)
  )

  enter-res
  )

; enter-duration-min prepares the arguments for the low level time-specialist
; routine add-duration-min

(defun enter-duration-min (a1 a2 d &aux
			       (pt1 (get-end a1))
			       (pt2 (get-start a2))
			       (dur d)
			       fn)
  (when (symbolp d)
        (add-interested-party d 'epilog::time-specialist)
	(setq fn (make-fn-arg 'epilog::min-of d))
        (setq dur (spec-eval-fn fn))
	(if (or (equal dur fn) (member dur '(unknown nil)))
	    (setq dur d))
	)
  (unless (symbolp dur)
          (add-duration-min pt1 pt2 dur)
          )
  )

; enter-duration-max prepares the arguments for the low level time-specialist
; routine add-duration-max

(defun enter-duration-max (a1 a2 d &aux
			       (pt1 (get-end a1))
			       (pt2 (get-start a2))
			       (dur d)
			       fn)
  (when (symbolp d)
        (add-interested-party d 'epilog::time-specialist)
	(setq fn (make-fn-arg 'epilog::max-of d))
        (setq dur (spec-eval-fn fn))
	(if (or (equal dur fn) (member dur '(unknown nil)))
	    (setq dur d))
	)
  (unless (symbolp dur)
          (add-duration-max pt1 pt2 dur)
          )
  )

; enter-reln enters the given relation (stem strict1 strict2) between the
; points or events a1 and a2, and the possibley null event a3. At this stage
; there are no absolute time arguments.
; It attempts to minimize the number of chains created by ordering the
; relations in a specific way, and using between rather than after/before
; where possible.
; enter-reln assumes all its arguments are events. If they are points this leads
; to some duplication of effort, but no inconsitencies.

(defun enter-reln (a1 stem a2 a3 strict1 strict2 &aux
		       (a1start (get-start a1))
		       (a2start (get-start a2))
		       (a3start (get-start a3))
		       (a1end (get-end a1))
		       (a2end (get-end a2))
		       (a3end (get-end a3)))
  (unless (null a3)
	  (enter-reln a2 'during a3 nil nil nil))

  ; handle individual relations

  (case stem
	((equal same-time)
	 (enter-point a1start 'equal a2start nil nil nil)
	 (enter-point a1end 'equal a2end nil nil nil)
	 )
	(before
	 (when (null a3)
	       (enter-point a1end 'before a2start nil strict1 nil)
	       (enter-point a1start 'before a1end nil nil nil)
	       )
	 (unless (null a3)
		 (enter-point a1end 'between a3start 
			      a2start nil strict1)
		 (enter-point a1start 'between a3start
			      a1end nil nil)
		 )
	 )
	(after
	 (when (null a3)
	       (enter-point a1start 'after a2end nil strict1 nil)
	       (enter-point a1end 'after a1start nil nil nil)
	       )
	 (unless (null a3)
		 (enter-point a1start 'between a2end 
			      a3end strict1 nil)
		 (enter-point a1end 'between a1end
			      a3end nil nil)
		 )
	 )
	(during
	 (unless (null a3)
		 (enter-point a2start 'after a3start nil nil nil)
		 (enter-point a2end 'before a3end nil nil nil)
		 )
	 (enter-point a1start 'between a2start a2end strict1 nil)
	 (enter-point a1end 'between a1start a2end nil strict2)
	 )
	(contains
	 (when (null a3)
	       (enter-point a1start 'before a2start nil strict1 nil)
	       (enter-point a1end 'after a2end nil strict2 nil)
	       )
	 (unless (null a3)
		 (enter-point a1start 'between a3start a2start nil strict1)
		 (enter-point a1end 'between a2end a3end strict2 nil)
		 )
	 (enter-point a1start 'before a1end nil nil nil)
	 )
	(overlaps
	 (when (null a3)
	       (enter-point a1end 'between a2start a2end nil strict2)
	       (enter-point a1start 'before a2start nil strict1 nil)
	       )
	 (unless (null a3)
		 (enter-point a1start 'between a3start a2start nil strict1)
		 (enter-point a1end 'between a2start a2end nil strict2)
		 (enter-point a2end 'before a3end nil nil nil)
		 )
	 )
	(overlapped-by
	 (when (null a3)
	       (enter-point a1start 'between a2start a2end strict1 nil)
	       (enter-point a1end 'after a2end nil strict2 nil)
	       )
	 (unless (null a3)
		 (enter-point a1start 'between a2start a2end strict1 nil)
		 (enter-point a1end 'between a2end a3end strict2 nil)
	         (enter-point a3start 'before a2start nil nil nil)
		 )
	 )
	)
  (if (event a2)
      (enter-point a2start 'before a2end nil nil nil)
      )
  )

; enter-absolute is used when a2 is an absolute time. The relation dictates
; whether it is to be used as an upper or lower bound on the absolute time of
; a1, or both. If a1 is also an absolute time, nothing is done.
; This routine also assumes that the other argument is an event - again, if
; the argument is a point, no inconsistencies will result (start or end of a
; point is the point itself)
; the a3 argument is ignored, although it probably shouldn't be.

(defun enter-absolute (a1 stem a2 a3 strict1 strict2 &aux
			   (abs2 (get-abs a2))
			   start end)
  (declare (ignore a3 strict1 strict2))
  ; a2 is absolute time
  ; if a1 is also, don't do anything
  (when (not (absolute-time a1))
	(setq start (get-start a1))
	(setq end (get-end a1))

	; handle individual relations

	(case stem
	      (before
	       (add-absolute-max end abs2))
	      (after
	       (add-absolute-min start abs2))
	      (equal
	       (add-absolute-min start abs2)
	       (add-absolute-max end abs2))
	      (during
	       nil)  ; cannot do anything if a during abs time
	      (contains
	       (add-absolute-max start abs2)
	       (add-absolute-min end abs2))
	      (overlaps
	       (add-absolute-min start abs2))
	      (overlapped-by
	       (add-absolute-max start abs2))
	      )
        (if (event a1)
            (enter-point start 'before end nil nil nil)
            )
	)
  )

; enter-abs-between handles a between relation when at least one of the
; arguments is an absolute time. If all are, nothing is done. Otherwise
; the appropriate absolute time bounds are set, or relation entered.

(defun enter-abs-between (a1 stem a2 a3 strict1 strict2)

  (declare (ignore stem))
  ; if a1 is not an absolute time, 
  ;   if a2 is, enter it as a1's minimum, otherwise enter the relation
  ;   if a3 is, enter it as a1's maximum, otherwise enter the relation

  (when (not (absolute-time a1))
	(if (absolute-time a2)
	    (add-absolute-min (get-start a1) (get-abs a2))
	    (enter-reln a1 'after a2 nil strict1 nil)
	    )
	(if (absolute-time a3)
	    (add-absolute-max (get-end a1) (get-abs a3))
	    (enter-reln a1 'before a3 nil strict2 nil)
	    )
	)

  ; if a1 is an absolute time,
  ;   if a2 is, nothing happens, otherwise a1 becomes a2's maximum
  ;   if a3 is, nothing happens, otherwise a1 becomes a3's minimum
  (when (absolute-time a1)
	(if (absolute-time a2)
	    nil
	    (add-absolute-max (get-end a2) (get-abs a1))
	    )
	(if (absolute-time a3)
	    nil
	    (add-absolute-min (get-start a3) (get-abs a1))
	    )
	)

  ; ensure that any event always has its start before its end

  (if (event a1)
      (enter-point (get-start a1) 'before (get-end a1) nil nil nil)
      )
  (if (event a2)
      (enter-point (get-start a2) 'before (get-end a2) nil nil nil)
      )
  (if (event a3)
      (enter-point (get-start a3) 'before (get-end a3) nil nil nil)
      )
  )

; enter-between handles the relation between when none of the arguments are
; absolute times (for relation between)

(defun enter-between (a1 stem a2 a3 strict1 strict2 &aux
			  (a1start (get-start a1))
			  (a1end (get-end a1))
			  (a2end (get-end a2))
			  (a3start (get-start a3)))
  (declare (ignore stem))
  (enter-point a1end 'between a2end a3start nil strict2)

  ; ensure that any event always has its start before its end

  (if (event a1)
      (enter-point a1start 'between a2end a1end strict1 nil)
      )
  (if (event a2)
      (enter-point (get-start a2) 'before a2end nil nil nil)
      )
  (if (event a3)
      (enter-point a3start 'before (get-end a3) nil nil nil)
      )
  )

