;; March 15th 2005 modified various functions to remove the -infinity
;; from durations. A duration cannot be negative. FM.

;;       Last change    August 1993
(in-package time)
;                 INFORMATION Routines

; find-psuedo returns the most strict relation it can between pt1 and pt2
; using their psuedo times

(defun find-psuedo (pt1 pt2 &aux
			(p1 (time-node-psuedo pt1))
			(p2 (time-node-psuedo pt2)))
  ; if the psuedo times are equal, they are the same point

  (cond ((equal p1 p2) 'same-time) ; 'equal

	; if the psuedo for pt1 < psuedo for pt2, then just have to determine
	; whether it is strict or not

	((< p1 p2) 
	 (if (possibly-equal pt1 pt2)
	     'before
	     'before-1)
	 )

	; similarly if psuedo for pt2 < psuedo for pt1

	((> p1 p2) 
	 (if (possibly-equal pt2 pt1)
	     'after
	     'after-1)
	 )

	; otherwise unknown (this should never happen - violates defn of chain)

	(t 'unknown)
	)
  )

; combine-path returns the strictness value of combining two paths, one of
; strictness s1, the other of strictness s2 (s1,s2 may only be before, before-1)

(defun combine-path (s1 s2)
  (if (or (equal s1 'before-1)
          (equal s2 'before-1)
	  (equal s1 t)
	  (equal s2 t)
	  (strict s2))
      'before-1
      'before)
  )

; calc-path returns the strictness value resulting from adding path1 and link
; to sofar

(defun calc-path (sofar path1 link &aux
			(st (time-link-strict link)))
  (combine-path sofar (combine-path path1 st))
  )

  
; there are two implicit links involved in a link between chains  
;    1) between the point starting from (pt1) on the first chain, 
;       and the xchain link from point, 
;    2) between the link to point on the second chain and the second 
;       point (pt2) 
; these can be < or <= and so must be included in path
;
;                      to----------->pt2
;                    /   
;                   / x-chain link   
;                  /
;    pt1----->from

; check-chain checks to see if the item link given is usable - its to point
; is before pt2 on the same chain. If so, it returns the resulting strictness
; from going to pt2 after sofar (includes the link itself)


(defun check-chain (sofar pt2 item &aux
				   (path (find-psuedo
					  (time-point
					   (time-link-to-point-name item))
					  pt2)))
  (when (equal (test-point-answer 'before path) 'yes)
    	(combine-path sofar path)
	)
  )


; search-meta returns nil if there is no path from pt1 to pt2
;                     before-1 if there is a strict path
;                 and before if there is a non strict path
;  Any path cannot go through any chain in already
;  Sofar is the strictness value so far in the search
; If it finds a path, but the path is not strict, the search will continue,
; just in case there is another path which is strict. If a strict path is
; found, the search stops, as no more information can be gained by continuing.

(defun search-meta (pt1 pt2 already sofar &aux
			      (chain2 (time-node-chain pt2))
			      (chain1 (time-node-chain pt1))
			      xlist res newsofar saveres)
  (if (traced-p 'epilog::time-search)
      (print-line "    Time Specialist: Searching for path from " 
		       (time-node-name pt1) " to " (time-node-name pt2)) )
  (when (getp (time-node-name pt1) 'rel)
	(setq res (getp (time-node-name pt1) 'rel))
	(return-from search-meta res) )

  (unless (null chain1) (setq xlist (meta-node-connections chain1)))

  (if (or res (null xlist))
      nil

      ; for each connection that the chain of pt1 has to another chain:

      (dolist (item xlist)
	      (let* ((frompt (time-point (time-link-from-point-name item)))
	             (topt (time-point (time-link-to-point-name item)))
	             (path1 (find-psuedo pt1 frompt))
	             (newchainno (to-chain-number item)))

	             ; see if this link is usable (must be before or equal pt1)

	             (when (equal (test-point-answer 'before path1) 'yes)
			   (setq newsofar (calc-path sofar path1 item))

			   ; if we got the end chain, see if this ends the srch
			   ; (to point must be before or equal pt2)

		           (if (equal newchainno
				      (meta-node-chain-number chain2))
			       (setq res (check-chain newsofar pt2 item))

			       ; otherwise continue search if this chain hasnt
			       ; been searched yet

                               (unless (member newchainno already)

                                       ; search for path from the end point 
			               ; of the link to the second point

				       (setq res
                                             (search-meta topt pt2 
	                                                  (adjoin newchainno
								  already)
							  newsofar))
				       )
			       )

			   ; if we have an answer, return it - else continue
			   ; with next connection from pt1's chain 

	                   (unless (or (null res) (equal res 'unknown))

				   ; if we have a strict path return it
				   ; if nonstrict, save it and continue search

				   (if (equal res 'before-1)
	                               (return res)
				       (let ()
				            (setq saveres res)
					    (setq res nil)
					    ))
	                           )
		           ))
	      ))

  ; if no answer, see if we saved one earlier

  (setq res (if (or (null res) (equal res 'unknown)) saveres res))
  (if (null res) (setq res 'unknown))
  (when res
	(push (time-node-name pt1) *changed-points*)
	(change-property (time-node-name pt1) 'rel res) )
  (if (equal res 'unknown) (setq res nil))
  res)

; compare-absolute-times returns the relation between pt1 and pt2 based
; on their absolute times. If the minimum of one is greater than the
; maximum of the other, the former is after the latter.
; If both absolute times are identical (min1 = min2, max1 = max2), the
; relation returned is equal - note however that the relation may
; actually be <= or >= depending on the presence or absence of paths.

(defun compare-absolute-times (pt1 pt2 &aux
				   (absmin1 (time-node-absolute-min pt1))
				   (absmax1 (time-node-absolute-max pt1))
				   (absmin2 (time-node-absolute-min pt2))
				   (absmax2 (time-node-absolute-max pt2))
				   (test1 (compare-abs absmax2 absmin1))
				   (test2 (compare-abs absmax1 absmin2)))

  ; if max of pt1 is before min of pt2, then pt1 is before pt2

  (if (equal (test-point-answer 'before test2) 'yes)
      (if (member test2 '(equal same-time))
	  'before
	  test2)

      ; if max of pt2 is before min of pt1, then pt1 is after pt2

      (if (equal (test-point-answer 'before test1) 'yes)
	  (if (member test1 '(equal same-time))
	      'after
	      (if (equal test1 'before)
		  'after
		  'after-1)
	      )

	  ; if min of pt1 = min of pt2 and max of pt1 = max of pt2
	  ; then they are equal
	  ; otherwise there is no way to tell the reln using absolute times

	  (let ((test3 (compare-abs absmin1 absmin2))
		(test4 (compare-abs absmax1 absmax2)))
               (if (and (equal (test-point-answer 'equal test3) 'yes)
	                (equal (test-point-answer 'equal test4) 'yes))
		   'same-time ; 'equal
	           'unknown
		   )
	       )
	  )
      )
  )

(defvar *changed-points* "keeps track of points with rels stuck on")

; search-path returns nil if there is no path from pt1 to pt2
;  or before-1 or before if there is

(defun search-path (pt1 pt2 &aux res)
  (setq *changed-points* nil)
  (setq res (search-meta pt1 pt2
			 (list (meta-node-chain-number
				(time-node-chain pt1)))
			 nil))
  (dolist (p *changed-points*) (remprop p 'rel))
  res)

; find-reln finds the most strict relation that holds between pt1 and pt2
; effort indicates how hard it should search

(defun find-reln (pt1 pt2 effort &aux
	              (result 'unknown)
		      (backup 'unknown))
  ; if on the same chain, compare psuedo times

  (when (equal pt1 pt2) (setq result 'same-time)) ; 'equal

  (when (on-same-chain pt1 pt2) (setq result (find-psuedo pt1 pt2)))

  ; if no answer yet, compare absolute times

  (when (equal result 'unknown)
	(setq result (compare-absolute-times pt1 pt2))

	(if (and (traced-p 'epilog::abs-time-eval)
		 (not (equal result 'unknown)))
	    (print-line "    Time Specialist: Found relation " 
			     (time-node-name pt1) " " result
			 " " (time-node-name pt2) " using absolute times"))

	; if the result is equal, there may still be a path indicating
	; a temporal order (<= or >=). Set result unknown so that this
	; path will be pursued, but save equal result just in case

	(when (and (member result '(equal same-time)) (> effort '0))
	    (setq backup result)
	    (setq result 'unknown)))

  ; if no answer yet, and effort indicates ok to continue, search for
  ; path from pt1 to pt2, or pt2 to pt1

  (when (and (equal result 'unknown) (> effort '0))
        (setq result
	      (cond ((search-path pt1 pt2))
		    ((inverse-reln (search-path pt2 pt1)))
		    (t 'unknown)
		    )) )

  ; if absolute time comparisons gave equal and the search gave no more 
  ; information, use the equal

  (if (and (equal result 'unknown) (not (equal backup 'unknown)))
      (setq result backup)
      )

  result
  )

; find-point-reln returns the most strict relationship that holds 
; between t1 and t2
; effort indicates how hard it should try - 0 - constant time operations only
;                                         >0   search

(defun find-point-reln (t1 t2 effort &aux
		            (result 'unknown))
  (if (equal t1 t2)
      (setq result 'same-time) ;'equal

      (if (or (absolute-time t1)
              (absolute-time t2))
          (setq result (find-absolute t1 t2 effort))
          (let ((pt1 (time-point t1))
	        (pt2 (time-point t2)))
	       (if (or (null pt1)
		       (null pt2))
	           (setq result 'unknown)
	           (setq result (find-reln pt1 pt2 effort))
	           )
	       )
          )
      )
  result
  )

; find-point returns the result given by find-point-reln and traces the
; call if desired

(defun find-point (t1 t2 effort &aux
		       (result (find-point-reln t1 t2 effort)))
  (when (traced-p 'epilog::time-point-eval)
	(print-line "    Time Specialist: Found point relation " t1 
			 " " result " " t2)
	)
  result
  )
;                 Duration Routines

; combine-durations adds the two durations together. If the min of either is
; nil, it is considered 0. If the max is, it is considered +infinity, and is
; carried through that way.

(defun combine-durations (d1 d2 &aux
			     (oldmin (car d1))
			     (oldmax (cadr d1))
			     (newmin (car d2))
			     (newmax (cadr d2))
			     calcmin calcmax)
  (if (null oldmin)
      (setq calcmin newmin)
      (if (null newmin)
	  (setq calcmin oldmin)
	  (setq calcmin (+ oldmin newmin))
	  )
      )

  (if (null oldmax)
      (setq calcmax nil)
      (if (null newmax)
	  (setq calcmax nil)
	  (setq calcmax (+ oldmax newmax))
	  )
      )

  (list calcmin calcmax)
  )

; get-best-duration returns the best duration between dur1 and dur2. Since both
; are assumed to be true for the same points, we can just pick the best 
; minimum (the maximum one) and the best maximum (the minimum one)

(defun get-best-duration (dur1 dur2)
  (let ((min1 (or (car dur1) 0))
	(min2 (or (car dur2) 0))
	(max1 (or (cadr dur1) 'epilog::+infinity))
	(max2 (or (cadr dur2) 'epilog::+infinity))
	)
    (assert (and (numberp min1) (numberp min2)))
    (if (not (numberp max1))
	(setq max1 max2))
    (if (not (numberp max2))
	(setq max2 max1))
    (list (max min1 min2)
	  (if (numberp max1)
	      (min max1 max2)
	    'epilog::+infinity))
    )
  )

; (defun get-best-duration (dur1 dur2)
;   (if (null dur1)
;       dur2
;     (if (null dur2)
; 	dur1
;       (let ((min1 (car dur1))
; 	    (min2 (car dur2))
; 	    (max1 (cadr dur1))
; 	    (max2 (cadr dur2))
; 	    newmin newmax)
; 	(if (or (null min1)
; 		(equal min1 'epilog::-infinity)) ;; this shouldn't be anymore possible
; 	    (setq newmin min2)
; 	  (if (or (null min2)
; 		  (equal min2 'epilog::-infinity))
; 	      (setq newmin min1)
; 	    (setq newmin (max-num min1 min2))
; 	    )
; 	  )
; 	(if (or (null max1)
; 		(equal max1 'epilog::+infinity))
; 	    (setq newmax max2)
; 	  (if (or (null max2)
; 		  (equal max2 'epilog::+infinity))
; 	      (setq newmax max1)
; 	    (setq newmax (min-num max1 max2))
; 	    )
; 	  )
; 	(list newmin newmax)
; 	)
;       )
;     )
;   )

; link-duration calculates the duration on a link, using the duration
; information stored there, and information gained from the absolute times

(defun link-duration (link &aux
			   (pt1 (time-point (time-link-from-point-name link)))
			   (pt2 (time-point (time-link-to-point-name link)))
			   (absdur (abs-duration pt1 pt2))
			   (dmin (time-link-duration-min link))
			   (dmax (time-link-duration-max link))
			   )
  (get-best-duration absdur (list dmin dmax))
  )

; abs-duration determines the duration between two points based on their
; absolute times

(defun abs-duration (pt1 pt2)
  (let (min1 max1 min2 max2)
    (when pt1
      (setq min1 (time-node-absolute-min pt1))
      (setq max1 (time-node-absolute-max pt1)))
    (when pt2
      (setq min2 (time-node-absolute-min pt2))
      (setq max2 (time-node-absolute-max pt2)))
    (list (calc-duration-min max1 min2)
	  (calc-duration-max min1 max2)))
  )

; search-for-duration returns nil if there is no path from pt1 to pt2
; and a list of the minimum and maximum durations if there is
;  Any path cannot go through any point in already

(defun search-for-duration (pt1 pt2 dur already &aux
				(desclist (append (time-node-descendants pt1)
						  (time-node-xdescendants pt1)))
				usedur curdur)
  (if (and (traced-p 'epilog::time-search)
	   (traced-p 'epilog::time-duration-eval))
      (print-line "    Time Specialist: Searching for duration from " 
		       (time-node-name pt1) " to " (time-node-name pt2)) 
      )
  (if (null desclist) 
      (setq usedur nil)

      ; go through all descendants (in chain and cross chain) of pt1

      (dolist (item desclist)
	      (let* ((topt (time-point (time-link-to-point-name item)))
		     (linkdur (link-duration item)))

		    ; make sure if thers is a loop, we dont loop

		    (unless (member (time-node-name topt) already)

                            (if (and (traced-p 'epilog::time-search)
				 (traced-p 'epilog::time-duration-eval))
		                (print-line
				 "    Time Specialist: Using link between " 
					   (time-node-name pt1) " and "
					   (time-node-name topt) 
					   *nl*
					   "       with duration " 
					   (return-duration linkdur)))

			    ; if this is the link, set the duration to this
			    ; link's duration. Otherwise add the new to dur

			    (if (null dur)
				(setq curdur linkdur)
			        (setq curdur (combine-durations dur linkdur)))

			    ; if this is the end point we're looking for
			    ; get the best duration of the ones so far

			    (if (equal topt pt2)
				(setq usedur (get-best-duration usedur curdur))

				; otherwise add this link and continue

				(setq usedur (get-best-duration usedur
				    (search-for-duration topt pt2 curdur
				      (append already 
					      (list (time-node-name topt))))))
				)
			    )
		    )
	      )
      )
  usedur
  )

;; return duration replaces nils in the duration list with what they
;; mean and reverse dmin and dmax if dmin is greater than dmax
(defun return-duration (dur &aux
			    (dmin (car dur))
			    (dmax (cadr dur)))
  (if (null dmin)
      (setq dmin 0)) ;; 0 was 'epilog::-infinity
  (if (null dmax)
      (setq dmax 'epilog::+infinity))
  (if (and (numberp dmin)
	   (numberp dmax)
	   (> dmin dmax))
      (list dmax dmin))
  (list dmin dmax)
  )

;; calc-duration determines the duration between two points.  If
;; either point doesn't exist, unknown If they exist, it first
;; determines the duration based on their absolute times. If the
;; minimum is greater than the max, the points are reversed.  If we
;; have a range, and the effort level (effort) indicates to continue
;; trying, search-for-duration is called to determine the best
;; duration along any path. The best between this and the absolute
;; time duration is returned.

(defun calc-duration (t1 t2 effort &aux
				   (pt1 (time-point t1))
				   (pt2 (time-point t2))
				   )
  (if (or (null pt1)
	  (null pt2))
      (progn
	(if (traced-p 'epilog::time-duration-eval)
	    (print-line "    Time Specialist WARNING: One of the time points is NIL. Found duration between "
			(print-info t1) " and "
			(print-info t2) " " 0 " " '+infinity)) ;; 0 was '-infinity
	;; 0 was 'epilog::-infinity but a duration can have value
	;; 'epilog::-infinity. FM
	(list 0 'epilog::+infinity)
	)
    (let* ((durans (abs-duration pt1 pt2))
	   (durmin (car durans))
	   (durmax (cadr durans)))
      (if (and (or (null durmin)
		   (equal durmax 'epilog::+infinity)
		   (null durmax))
	       (> effort 0))
	  (setq durans
	    (get-best-duration
	     durans
	     (search-for-duration pt1 pt2 nil (list (time-node-name pt1)))))
	)
      
      ;; replace nils in the result
      (setq durans (return-duration durans))
      
      (if (traced-p 'epilog::time-duration-eval)
	  (print-line "    Time Specialist: Found duration between " 
		      (print-info t1) " and "
		      (print-info t2) " " (car durans) " " (cadr durans)))
      durans)
    )
  )


; find-relation returns the most strict relation it can find between a1 and a2
; It assumes it is worjing with events (again - no problems with points)
; It determines relationships between the starts, ends , and start of one, end
; of other and uses those results to determine the actual realtion.
; It attempts to minimize the number of comparsions it does.

(defun find-relation (a1 a2 effort)
  (if (equal a1 a2)
      'same-time ;'equal

  (if (or (missing a1)
	  (missing a2))
      'unknown

  (let* ((a1start (get-start a1))
	 (a1end (get-end a1))
	 (a2start (get-start a2))
	 (a2end (get-end a2))
	 (result 'unknown)
	 (e1s2 (find-point a1end a2start effort))
	 s1e2 s1s2 e1e2 strict1 strict2)

  ; if the end of a1 is before the start of a2, a1 is before a2

  (when (or (equal (test-point-answer 'before e1s2) 'yes)
	    (and (not (event a1))
		 (not (event a2))))
	(if (and (member e1s2 '(equal same-time))
		 (or (event a1)
		     (event a2)))
	    (setq result 'before-0)
            (setq result e1s2)))

  ; if the start of a1 is after the end of a2, a1 is after a2

  (when (and (equal result 'unknown)
	     (or (event a1)
		 (event a2)))
	(setq s1e2 (find-point a1start a2end effort))
	(if (equal (test-point-answer 'after s1e2) 'yes)
	    (if (and (member s1e2 '(equal same-time))
		     (or (event a1)
		         (event a2)))
	        (setq result 'after-0)
	        (setq result s1e2)))
	)

  ; if the start points are equal, and the end points are equal, a1 = a2

  (when (and (equal result 'unknown)
	(or (event a1)
	    (event a2)))
	(setq s1s2 (find-point a1start a2start effort))
	(setq e1e2 (find-point a1end a2end effort))
	(if (and (equal (test-point-answer 'equal s1s2) 'yes)
		 (equal (test-point-answer 'equal e1e2) 'yes))
	    (setq result 'same-time)) ;'equal
	)

  ; all other relations require that at least one of the arguments be an event

  (when (and (equal result 'unknown)
	     (or (event a1)
		 (event a2)))
	(setq strict1 (first-strict (split-time-pred s1s2)))
	(setq strict2 (first-strict (split-time-pred e1e2)))
	(when (member s1s2 '(equal same-time))
	      (setq strict1 '0))
	(when (member e1e2 '(equal same-time))
	      (setq strict2 '0))

	; if the start of the first is after the start of the second
	; a2.start    a1.start
	; a1 is either during a2, or overlapped by it

	(when (equal (test-point-answer 'after s1s2) 'yes)

	      ; if the end of e1 is before the end of e2 - during
	      ; a2.start  a1.start  a1.end  a2.end

	      (when (equal (test-point-answer 'before e1e2) 'yes)
	            (setq result (build-pred 'during strict1 strict2))
		    )

	      ; if the end of e1 is after the end of e2 - overlapped-by
	      ; a2.start  a1.start  a2.end  a1.end

	      (when (equal (test-point-answer 'after e1e2) 'yes)
	            (setq result (build-pred 'overlapped-by strict1 strict2))
		    )
	      )

	; if the start of the first is before the start of the second
	; a1.start    a2.start
	; a1 either contains a2, or overlaps it

	(when (equal (test-point-answer 'before s1s2) 'yes)

	      ; if the end of e1 is before the end of e2 - overlaps
	      ; a1.start  a2.start  a1.end  a2.end

	      (when (equal (test-point-answer 'before e1e2) 'yes)
	            (setq result (build-pred 'overlaps strict1 strict2))
		    )

	      ; if the end of e1 is after the end of e2 - contains
	      ; a1.start  a2.start  a2.end  a1.end

	      (when (equal (test-point-answer 'after e1e2) 'yes)
	            (setq result (build-pred 'contains strict1 strict2))
		    )
	      )
	)
  (when (traced-p 'epilog::time-eval)
      (print-line "    Time Specialist: Found relation " (print-info a1)
		       " " result " " 
		       (print-info a2)) 
      )
  result)))
  )


; abs-relation determines the relation between an abs time and argument arg

(defun abs-relation (abs arg &aux
			 (pt (time-point arg)))
  (if (null pt)
      'unknown
      (let* ((res1 (compare-abs abs
				(time-node-absolute-min pt)))
	     (res2 (compare-abs abs
				(time-node-absolute-max pt))))
            (cond ((and (equal (test-point-answer 'equal res1) 'yes)
	                (equal (test-point-answer 'equal res2) 'yes))
                   'same-time) ;'equal
                  ((equal (test-point-answer 'before res1) 'yes)
		   (if (member res1 '(equal same-time))
		       'before
	                res1)
		   )
	          ((equal (test-point-answer 'after res2) 'yes)
		   (if (member res2 '(equal same-time))
		       'after
	                res2)
		   )
	          (t 'unknown)
	          )
	    )
      )
  )
	        
; find-absolute returns the relationship between a1 and a2, where one is an
; absolute time.

(defun find-absolute (a1 a2 effort &aux res)
  (setq res
  (if (absolute-time a1)
      (if (absolute-time a2)
	  (compare-abs (get-abs a1) (get-abs a2))
	  (abs-relation (get-abs a1) a2)
	  )
      (if (absolute-time a2)
	  (inverse-reln (abs-relation (get-abs a2) a1))
	  (find-point a1 a2 effort) ; just in case called without abs arg
	  )
      ))
  (if (or (traced-p 'epilog::abs-time-eval)
          (traced-p 'epilog::time-point-eval))
      (print-line "    Time Specialist: Found relation " (print-info a1 )
		       " " res " " (print-info a2) " using absolute times"))
  res
  )

(defun find-absolute-reln (a1 a2 effort &aux
			       (ans 'unknown))
  (setq ans
  (if (and (absolute-time a1)
	   (absolute-time a2))
      (find-absolute a1 a2 effort)
      (let* ((a1start (abs-start a1))
	     (a2start (abs-start a2))
	     (a1end (abs-end a1))
             (a2end (abs-end a2))
	     (res1 (find-absolute a1start a2end effort))
	     (res2 (find-absolute a1end a2start effort)))

	   ; if start and end are equal, equal

	    (cond ((and (equal (test-point-answer 'equal res1) 'yes)
	 	        (equal (test-point-answer 'equal res2) 'yes))
		   'same-time) ;'equal

		  ; if start of 1 after end of 2, after

		  ((equal (test-point-answer 'after res1) 'yes)
		   (if (member res1 '(equal same-time))
		       'after
		       res1)
		   )

		  ; if end of 1 before start of 2, before

		  ((equal (test-point-answer 'before res2) 'yes)
		   (if (member res2 '(equal same-time))
		       'before
		       res2)
		   )

		  (t 'unknown)
		  )
	   )
      ))
  (if (or (traced-p 'epilog::time-eval)
	  (traced-p 'epilog::abs-time-eval))
      (print-line  "    Time Specialist: Found relation " (print-info a1 )
		       " " ans " " (print-info a2)))
  ans

  )
