;       Last change    August 1993
(in-package time)
;              Temporal Predicate Splitting and Recombining

; Some of the temporal predicates have one or two strictness values
; embedded in them (eg. before-1, between-0-1). 
;     strictness value            meaning
;          0                         0 space between   (meets)
;          1                         >0 space between (strict < or >)
;         nil                        0 or 1  (<= or >= depending on stem)
;    
; These routines separate the stem from the strictness values, 
; and determine inverse relations for those temporal predicates. 


; split-time-pred divides the temporal predicate into  a stem and strictness
; indicators. Each predicate looks like stem[-s1[-s2]]
; The strictness indicators may be nil, 0 or 1. If the predicate ends in nil,
; the preceding - is dropped.
; Eg's before
;      before-1
;      before-0-1
;      before--1
;
; returns a list of the form (stem strict1 strict2)

(defun split-time-pred (pred &aux
			(stem nil)
			(s1 nil)
			(s2 nil)
			(reln (strip-pkg pred))
			pos newpos)
  ; first get the stem
  (if (null pred)
      (setq pos nil)
      (setq pos (position '- reln)))
  (if (null pos)
      (setq stem pred)
      (let ()
	   ; check for special case overlapped-by

	   (setq stem (pack (subseq reln '0 pos)))
	   (when (member stem '(overlapped epilog::overlapped
			        same epilog::same
				has epilog::has))
		 (setq pos (+ pos 1))
		 (setq pos (position '- reln :start pos))
		 (if (null pos)
		     (setq stem pred)
		     (setq stem (pack (subseq reln '0 pos)))
		     )
		 )
	   )
      )

  ; get strictness values if there are any

  (unless (null pos)
	  (setq pos (+ pos '1))
	  (setq newpos (position '- reln :start pos))
	  (if (null newpos)
	      (setq s1 (car (subseq reln pos)))
	      (let ()
		   (setq s1 (car (subseq reln pos newpos)))
		   (setq pos (+ newpos '1))
		   (setq newpos (position '- reln :start pos))
		   (if (null newpos)
		       (setq s2 (car (subseq reln pos)))
		       (setq s2 (car (subseq reln pos newpos)))
		       )
		   )
	      )
	  )

  ; make sure symbols returned are local to time

  (unless (null stem)
	  (setq stem (find-symbol (symbol-name stem) 'time)))
  (unless (null s1)
	  (setq s1 (find-symbol (symbol-name s1) 'time)))
  (unless (null s2)
	  (setq s2 (find-symbol (symbol-name s2) 'time)))
  (list stem s1 s2)
  )

; Stem-of returns the first element of a list returned by split-time-pred

(defmacro stem-of (plist)
  `(car ,plist)
  )

; first-strict returns the second element of a list returned by split-time-pred

(defmacro first-strict (plist)
  `(cadr ,plist)
  )

; second-strict returns the third element of a list returned by split-time-pred

(defmacro second-strict (plist)
  `(caddr ,plist)
  )

; build-pred builds a predicate symbol given the stem and strictness values
; eg before + strict   =>  before-1
;    curing + strict + meets  =>  during-1-0

(defun build-pred (stem &optional strict1 strict2 &aux
			(usestem (pack (strip-pkg stem)))
			(s1 strict1)
			(s2 strict2))
  (when (equal strict1 '|1|)
        (setq s1 '1))
  (when (equal strict1 '|0|)
        (setq s1 '0))
  (when (equal strict2 '|1|)
        (setq s2 '1))
  (when (equal strict2 '|0|)
        (setq s2 '0))
  (if (null s1)
      (if (null s2)
	  stem
	  (find-symbol (symbol-name (pack (list usestem '- '- s2))) 'time)
	  )
      (if (null s2)
	  (find-symbol (symbol-name (pack (list usestem '- s1))) 'time)
	  (find-symbol (symbol-name (pack (list usestem '- s1 '- s2))) 'time)
	  )
      )
  )

; more-strict returns the strictest version of the given pred

(defun more-strict (pred &aux
			 (plist (split-time-pred pred))
			 (stem (stem-of plist)))
  (case stem 
	((equal at exactly) pred)
	((before after) 
	 (if (null (first-strict plist))
	     (build-pred stem '1 nil)
	     pred))
	(t (apply 'build-pred (subst '1 nil plist)))
	)
  )

; inverse-stem returns the inverse operation for stem
; eg. the inverse of before is after

(defun inverse-stem (stem)
   (case stem 
	 (before 'after)
	 (after 'before)
	 (during 'contains)
	 (contains 'during)
	 (overlaps 'overlapped-by)
	 (overlapped-by 'overlaps)
	 (otherwise 'unknown)  ; equal between unknown
	 )
  )

; inverse-reln returns the inverse relation of reln. Only the stem
; needs to be inverted, as the strictness values are based on the order
; of the points.

(defun inverse-reln (reln &aux
			  (plist (split-time-pred reln))
			  (stem (stem-of plist))
			  (s1 (first-strict plist))
			  (s2 (second-strict plist)))
  (unless (null reln)
	  (build-pred (inverse-stem stem) s1 s2)
	  )
  )

; Strict returns t is the strictness value is strict, nil otherwise

(defmacro strict (x)
  `(if (or (equal ,x '|1|)
	   (equal ,x '1))
       t
       nil)
  )

; Notstrict returns t if the strictness value is not strict, nil otherwise

(defmacro notstrict (x)
  `(null ,x)
  )

; Meets returns t if the strictness value indicates exact meet, nil otherwise

(defmacro meets (x)
  `(if (or (equal ,x '|0|)
	   (equal ,x '0))
       t
       nil)
  )

; Combine-Strict  - combines the strictness values of two arcs to 
; determine what a combined arc would be. 
; < + anything  -> <
; anything + <  -> <
; x + x         -> x  (if same, combo is same)
; otherwise     -> <=

(defun combine-strict (strict1 strict2)
  (cond ((equal strict1 strict2)
         strict1)
	((strict strict1)
	 strict1)
	((strict strict2)
	 strict2)
	((meets strict1)
	 strict2)
	((meets strict2)
	 strict1)
	(t nil)
	)
  )

;                Comparison Routines

; compare-strict compares the two strictness values for the same stem and
; decides what relationship would hold between the predicates.

(defun compare-strict (s1 s2)
  (cond ((equal s1 s2)
	 'equivalent)
	((notstrict s1)
	 'subsumes)
	((notstrict s2)
	 'subsumed)
	(t 'disjoint)
	)
  )

; compare-inverse-strict does the same thing for stems which are the inverse
; of each other

(defun compare-inverse-strict (s1 s2)
  (cond ((or (strict s1) (strict s2))
	 'disjoint)
	((and (meets s1) (meets s2))
	 'equivalent)
	(t 'intersects)
	)
  )

; compare-single compares two predicates which may each have one strictness
; value associated with them (equal, before, after)

(defun compare-single (stem1 s1 stem2 s2)
  (cond ((and (member stem1 '(equal same-time))
	      (member stem2 '(equal same-time)))
	 'equivalent)
	
	((equal stem1 stem2)
         (compare-strict s1 s2))

	; for points, after-0 and before-0 are the same thing, also
	; after and before intersects.
	; For events or combos, not true

	((equal stem1 (inverse-stem stem2))
	 (compare-inverse-strict s1 s2))

	; for points, after-0 and before-0 are equivalent to equal
	; but for events or combos, they are not (subsumes?)

	((member stem1 '(equal same-time))
	 (if (meets s2)
	     'equivalent
	     (if (strict s2)
	         'disjoint
		 'subsumed)))
	((member stem2 '(equal same-time))
	 (if (meets s1)
	     'equivalent
	     (if (strict s1)
	         'disjoint
		 'subsumes)))
	(t 'unknown)
	)
  )

; determine-split determines what the predicates involved are when splitting
; a complex relation (during, between, contains, overlaps or overlapped-by)
; into several simple relations

(defun determine-split (stem1 s1 s2)
  (case stem1
	((equal same-time)
	 (list (list stem1 nil) 
	       (list stem1 nil) 
	       (list 'before nil) 
	       (list 'after nil))
	 )
	(during 
	 (list (list 'after s1) 
	       (list 'before s2) 
	       (list 'before (combine-strict nil s2)) 
	       (list 'after (combine-strict s1 nil)))
	 )
	(contains
	 (list (list 'before s1) 
	       (list 'after s2) 
	       (list 'before (combine-strict s1 nil)) 
	       (list 'after (combine-strict nil s2)))
	 )
	(overlaps
	 (list (list 'before s1) 
	       (list 'before s2) 
	       (list 'before (combine-strict s1 s2)) 
	       (list 'after nil))
	 )
	(overlapped-by
	 (list (list 'after s1) 
	       (list 'after s2) 
	       (list 'before nil)
	       (list 'after (combine-strict s1 s2)))	 
	 )
	(between
	 (list (list 'after s1)
	       (list 'before s2))
	 )
	(before
	 (list (list 'before (combine-strict nil s1))
	       (list 'before (combine-strict s1 nil)) 
	       (list 'before (combine-strict s1 nil)) 
	       (list 'before s1))
	 )
	(after
	 (list (list 'after (combine-strict s1 nil)) 
	       (list 'after (combine-strict nil s1) )
	       (list 'after s1) 
	       (list 'after (combine-strict nil s1)))
	 )
	)
  )

(defun compare-results (res res1)
  (cond ((or (null res)
	     (equal res res1))
	 res1)
        ((equal res1 'disjoint)
         'disjoint)
	((equal res 'equivalent)
	 res1)
	((equal res1 'equivalent)
	 res)
        ((or (and (equal res 'subsumes)
                  (equal res1 'subsumed))
	     (and (equal res 'subsumed)
	          (equal res1 'subsumes)))
	 'intersects)
	((or (equal res 'intersects)
	     (equal res1 'intersects))
	 'intersects)
	(t 'unknown)
	)
  )

; split-and-compare uses determine-split to split the relations requested
; and uses them to determine the relationship between the given predicates.

(defun split-and-compare (stem1 s1 s2 stem2 t1 t2 &aux
				(sp1 (determine-split stem1 s1 s2))
				(sp2 (determine-split stem2 t1 t2))
				res)
  (dolist (item sp1)
	  (let* ((item2 (car sp2))
		 (res1 (compare-single (car item) (cadr item) (car item2)
		       (cadr item2))))
		(setq sp2 (cdr sp2))
		(setq res (compare-results res res1))
		(if (equal res 'disjoint)
		    (return res))
		)
	  )
  res
  )

; compare-equal - one of the stems is equal, the other before or after
; since this may deal with points or events, we must take the least restrictive
; of the two (events) 
; with points, before-0 and after-0 would be equivalent to equal, but for
; events, they subsume equal
; the split and compare used with the other predicates cannot be used here,
; because after and before imply relations between different points of the
; event than equal does
; equal (start1 = start2) (end1 = end2)
; before or after (start1 cp end2) (start2 cp end1)

; This function doesn't appear to be called anywhere.
; (defun compare-equal (stem1 s1 stem2 t1)
;   (if (member stem1 '(equal same-time))
;       (if (strict t1)
; 	  'disjoint
; 	  'subsumed)
;       ; else stem2 is equal
;       (if (strict s1)
; 	  'disjoint
; 	  'subsumes)
;       )
;   )

; compare-pred-parts compares two predicates which have been split into
; stems and strictness values
; the relationship between the predicates is returned.
	
(defun compare-pred-parts (stem1 s1 s2 stem2 t1 t2)
  (cond	((and (member stem1 '(between))
	      (member stem2 '(between)))
	 (split-and-compare stem1 s1 s2 stem2 t1 t2))
	((and (not (member stem1 '(between)))
	      (not (member stem1 '(between))))
	 (split-and-compare stem1 s1 s2 stem2 t1 t2))
	(t 'unknown)
	)
  )


; compare-time-preds returns the relationship that holds between pred1 and pred2
;    equivalent
;    subsumes
;    subsumed
;    disjoint
;    intersects
;    unknown

(defun compare-time-preds (pred1 pred2 &aux
		            (plist1 (split-time-pred pred1))
		            (plist2 (split-time-pred pred2)))
  (compare-pred-parts (stem-of plist1) (first-strict plist1) 
		      (second-strict plist1)
		      (stem-of plist2) (first-strict plist2)
		      (second-strict plist2))
  )

; The next routines decide whether the requested relation and the actual
; relation are compatible (true)
;              opposite (false)
;              or unknown
; This involves testing not just the stems of the predicates, but their
; strictness values as well.

; test-result checks to see if result (result r1 r2) is acceptable for
; the testing relation (reln s1 s2) by comparing the stems and strictness
; values

(defun test-result (test s1 s2 result r1 r2 &aux
                         (comp (compare-pred-parts test s1 s2 result r1 r2)))
  (cond ((member comp '(equivalent subsumes))
         'yes)
        ((member comp '(disjoint))
         'no)
        ((member comp '(intersects subsumed))
         'unknown)
        (t 'unknown)
        )
  )

; test-point-result is that same as test-result, except that the relations are
; assumed to be between points only

(defun test-point-result (test s1 result r1 &aux
                               (comp (compare-single test s1 result r1)))
  (cond ((member comp '(equivalent subsumes))
         'yes)
        ((member comp '(disjoint))
         'no)
        ((member comp '(intersects subsumed))
         'unknown)
        (t 'unknown)
        )
  )

; test-answer returns yes if result is acceptable for reln,
;                     no if it is not
;                 and unknown if it is not known whether it is acceptable or not
(defun test-answer (reln result &aux
                         (plist1 (split-time-pred reln))
                         (plist2 (split-time-pred result)))
  (if (equal result 'unknown)
      'unknown
      (test-result (stem-of plist1) (first-strict plist1) (second-strict plist1)                   (stem-of plist2) (first-strict plist2) (second-strict plist2)                   )
      )  
  )
 
; test-point-answer is that same as test-answer, except that the relations are
; assumed to be between points only

(defun test-point-answer (reln result &aux
                         (plist1 (split-time-pred reln))
                         (plist2 (split-time-pred result)))
  (if (equal result 'unknown)
      'unknown
      (test-point-result (stem-of plist1) (first-strict plist1)
                         (stem-of plist2) (first-strict plist2))
      )  
  )

; combine-test-results takes the result of a number of tests and determines
; the "and"ed result.

(defun combine-test-results (&rest tests)
  (cond ((member nil tests)
         'no)
        ((member 'no tests)
         'no)
        ((member 'unknown tests)
         'unknown)
        (t 'yes)
        )
  )

